/*
  クリークファインディング
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "simcomp.h"

/* スタティック変数 */
static int  _clifinder_max_repeat_;  /* 最大リピート数 */
static int  _clifinder_repeat_;      /* リピート数の現在値 */
static MCS_CTL *mcslist;             /* 個別のクリーク情報 */
static unsigned int **connected;
static unsigned int *selected;
static int degree;
static int maxsize;
static int repeat_changeover;
static int nodenum;            /* ドッキングノード数 */
static int columnnum;
static unsigned int *maskbit;
#ifdef DEBUG
static SIMCOMP *psimcomp;
#endif

/* private function */
int explore_sum(int *, int);
int explore_numonly(int *, int);
void sort_(int, int *, int *);

void clifinder2(SIMCOMP *simcomp){
  int i;
  int *new;

  _clifinder_max_repeat_ = (simcomp->maxrepeat < 0) ? MAX_REPEAT : simcomp->maxrepeat;
  _clifinder_repeat_ = 0;
  mcslist = &(simcomp->mcslist);
  degree = 0;
  maxsize = 0;
#ifdef DEBUG
  psimcomp = simcomp;
#endif
  nodenum = simcomp->nodenum;
  columnnum = simcomp->columnnum;
  repeat_changeover = nodenum / 2;
  if (nodenum == 0) return;

  new = (int *)malloc(sizeof(int) * nodenum);
  if (new == NULL){
    simcomp->status = 6;
    return;
  }
  connected = simcomp->matrix;
  maskbit = simcomp->bitset1;
  selected = simcomp->bitset2;

  for (i = 0; i < nodenum; i++) new[i] = i;
  for (i = 0; i < columnnum; i++){
    maskbit[i] = 0;
    selected[i] = 0;
  }

  simcomp->status = explore_sum(new, nodenum);
  simcomp->mcs_max = maxsize;
#ifdef DEBUG
if (psimcomp->debug_mode % 2 == 1){
printf("repeat(%d)\n", _clifinder_repeat_);
}
#endif

  free(new);

  return;
}

int explore_sum(int* old, int oldnum){
  int i, j;
#ifdef DEBUG
  int l;
#endif
  int *sum;
  unsigned int *vlist;  /* 確保後はMCS内で管理する */
  int *new;
  int newnum;

#ifdef DEBUG
if (psimcomp->debug_mode % 2 == 1){
printf("explore_sum (%d)( ", degree);
for (l = 0; l < columnnum; l++) printf("%08x ", selected[l]);
printf(")\n");
}
#endif
  /* matrixデータに対するマスクビットを作成 */
  for (i = 0; i < columnnum; i++) maskbit[i] = 0;
  for (i = 0; i < oldnum; i++) setBit(maskbit, old[i]);

  sum = (int *)malloc(sizeof(int) * oldnum);
  if (sum == NULL){
    return 6;
  }

  /* ループの中で毎回数えてmallocするよりも */
  new = (int *)malloc(sizeof(int) * oldnum);
  if (new == NULL){
    free(sum);
    return 6;
  }

  /* カウント */
  for (i = 0; i < oldnum; i++){
    sum[i] = 0;
    for (j = 0; j < columnnum; j++){
      sum[i] += bitCount(connected[old[i]][j] & maskbit[j]);
    }
  }
  /* sum でソート */
  sort_(oldnum, sum, old);
#ifdef DEBUG
if (psimcomp->debug_mode % 2 == 1){
printf("  sort");
for(l = 0; l < oldnum; l++){
printf(" %d:%d", old[l] + 1, sum[l]);
}
printf("\n");
}
#endif

  /* クリークの発見 */
  if (sum[oldnum - 1] == oldnum){
#ifdef DEBUG
if (psimcomp->debug_mode % 2 == 1){
printf("FIND0 : degree(%d) oldnum(%d)   maxsize(%d)\n", degree, oldnum, maxsize);
}
#endif
    /* クリークが見つかった時の処理  〜 ここから 〜 */
    _clifinder_repeat_++;
    if (degree + oldnum > maxsize){
      /* 最大クリークなのでそれまでのクリークを全部クリア */
      freelist_MCS(mcslist);
      maxsize = degree + oldnum;
    }
    if (degree + oldnum == maxsize){
      vlist = (unsigned int *)malloc(sizeof(unsigned int) * columnnum);
      if (vlist == NULL){
        free(sum);
        free(new);
        return 6;
      }

      for (i = 0; i < columnnum; i++) vlist[i] = maskbit[i] | selected[i];
#ifdef DEBUG
if (psimcomp->debug_mode % 2 == 1){
printf("setlist_MCS0( ");
for (l = 0; l < columnnum; l++) printf("%08x ", vlist[l]);
printf(")\n");
}
#endif

      if (setlist_MCS(mcslist, degree + oldnum, vlist, 0.0) == _NO_){
        free(sum);
        free(new);
        free(vlist);
        return 6;
      }
    }
    /* クリークが見つかった時の処理  〜 ここまで 〜 */
    free(sum);
    free(new);
    return 0;
  }

  /* oldnum の更新 */
  for (i = 0; i < oldnum; i++){
    if (sum[i] + 1 < maxsize - degree){  /* sumには +1 が必要では？ */
      oldnum = i;
    }
  }
#ifdef DEBUG
if (psimcomp->debug_mode % 2 == 1){
printf("  oldnum(%d) update\n", oldnum);
}
#endif

  /* 一つ取っては繰り返し */
  for (i = 0; i < oldnum; i++){
    if (_clifinder_max_repeat_ &&
        ++_clifinder_repeat_ > _clifinder_max_repeat_){
#ifdef DEBUG
if (psimcomp->debug_mode % 2 == 1){
printf("OVER max repeat(%d)\n", _clifinder_max_repeat_);
}
#endif
      free(sum);
      free(new);
      return 0;
    }

#ifdef DEBUG
if (psimcomp->debug_mode % 2 == 1){
printf("oldnum(%d) sum[%d](%d)    maxsize(%d) degree(%d)\n", oldnum, i, sum[i], maxsize, degree);
}
#endif
    if (oldnum - i < maxsize - degree || sum[i] + 1 < maxsize - degree) break;  /* sumには +1 が必要では？ */

    newnum = 0;
    for (j = i + 1; j < oldnum; j++)
      if (getBit(connected[old[i]], old[j]) != 0) new[newnum++] = old[j];
#ifdef DEBUG
if (psimcomp->debug_mode % 2 == 1){
printf(" old[%d](%d)  newnum(%d)\n", i, old[i] + 1, newnum);
}
#endif

    if (newnum == 0){
      /* クリークが見つかった時の処理  〜 ここから 〜 */
#ifdef DEBUG
if (psimcomp->debug_mode % 2 == 1){
printf("FIND1 : degree(%d)   maxsize(%d)\n", degree, maxsize);
}
#endif
      _clifinder_repeat_++;
      if (degree + 1 > maxsize){
        /* 最大クリークなのでそれまでのクリークを全部クリア */
        freelist_MCS(mcslist);
        maxsize = degree + 1;
      }
      if (degree + 1 == maxsize){
        vlist = (unsigned int *)malloc(sizeof(unsigned int) * columnnum);
        if (vlist == NULL){
          free(sum);
          free(new);
          return 6;
        }
        for (j = 0; j < columnnum; j++) vlist[j] = selected[j];
        setBit(vlist, old[i]);  /* 追加するのはi番目 */

#ifdef DEBUG
if (psimcomp->debug_mode % 2 == 1){
printf("setlist_MCS1( ");
for (l = 0; l < columnnum; l++) printf("%08x ", vlist[l]);
printf(")\n");
}
#endif
        if (setlist_MCS(mcslist, degree + 1, vlist, 0.0) == _NO_){
          free(sum);
          free(new);
          free(vlist);
          return 6;
        }
      }
      /* クリークが見つかった時の処理  〜 ここまで 〜 */
    } else {
      /* 再帰的処理 */
      if (newnum + 1 >= maxsize - degree){
        setBit(selected, old[i]);
        degree++;
        if (_clifinder_repeat_ < repeat_changeover){
          explore_sum(new, newnum);
        } else {
          explore_numonly(new, newnum);
        }
        resetBit(selected, old[i]);
        degree--;
      }
    }
  }

  free(sum);
  free(new);

  return 0;
}

int explore_numonly(int* old, int oldnum){
  int i, j;
#ifdef DEBUG
  int l;
#endif
  int *num;
  int maxnum;
  unsigned int *vlist;  /* 確保後はMCS内で管理する */
  int *new;
  int newnum;

#ifdef DEBUG
if (psimcomp->debug_mode % 2 == 1){
printf("explore_numonly (%d)(%d)(%d)( ", degree, oldnum, maxsize);
for (l = 0; l < columnnum; l++) printf("%08x ", selected[l]);
printf(")");
}
#endif
  num = (int *)malloc(sizeof(int) * oldnum);
  if (num == NULL){
    return 6;
  }

  new = (int *)malloc(sizeof(int) * oldnum);
  if (new == NULL){
    free(num);
    return 6;
  }

  /* カラーリング */
  maxnum = 0;
  for(i = 0; i < oldnum; i++){
    for (j = 0; j < oldnum; j++) new[j] = 0;
    for (j = 0; j < i; j++)
      if (getBit(connected[old[i]], old[j]) != 0) new[num[j]] = 1;
    for (j = 0; j < oldnum; j++){
      if (new[j] == 0){
        if (maxnum < j) maxnum = j;
        break;
      }
    }
    num[i] = j;
  }

#ifdef DEBUG
if (psimcomp->debug_mode % 2 == 1){
printf(" (%d)\n", maxnum);
}
#endif
  /* 終了条件のチェック */
  if (maxnum + 1 < maxsize - degree){
    free(num);
    free(new);
    return 0;
  }

  /* num でソート */
  sort_(oldnum, num, old);
#ifdef DEBUG
if (psimcomp->debug_mode % 2 == 1){
printf("  sort");
for(l = 0; l < oldnum; l++){
printf(" %d:%d", old[l] + 1, num[l]);
}
printf("\n");
}
#endif

  /* 一つ取っては繰り返し */
  for(i = 0; i < oldnum; i++){
    if (_clifinder_max_repeat_ &&
        ++_clifinder_repeat_ > _clifinder_max_repeat_){
#ifdef DEBUG
if (psimcomp->debug_mode % 2 == 1){
printf("OVER max repeat(%d)\n", _clifinder_max_repeat_);
}
#endif
      free(num);
      free(new);
      return 0;
    }

    if (oldnum - i < maxsize - degree || num[i] + 1 < maxsize - degree) break;

    newnum = 0;
    for (j = i + 1; j < oldnum; j++)
      if (getBit(connected[old[i]], old[j]) != 0) new[newnum++] = old[j];

    if (newnum == 0){
      /* クリークが見つかった時の処理  〜 ここから 〜 */
#ifdef DEBUG
if (psimcomp->debug_mode % 2 == 1){
printf("FIND1 : degree(%d)   maxsize(%d)\n", degree, maxsize);
}
#endif
      if (degree + 1 > maxsize){
        /* 最大クリークなのでそれまでのクリークを全部クリア */
        freelist_MCS(mcslist);
        maxsize = degree + 1;
      }
      if (degree + 1 == maxsize){
        vlist = (unsigned int *)malloc(sizeof(unsigned int) * columnnum);
        if (vlist == NULL){
          free(num);
          free(new);
          return 6;
        }
        for (j = 0; j < columnnum; j++) vlist[j] = selected[j];
        setBit(vlist, old[i]);  /* 追加するのはi番目 */

#ifdef DEBUG
if (psimcomp->debug_mode % 2 == 1){
printf("setlist_MCS2( ");
for (l = 0; l < columnnum; l++) printf("%08x ", vlist[l]);
printf(")\n");
}
#endif
        if (setlist_MCS(mcslist, degree + 1, vlist, 0.0) == _NO_){
          free(num);
          free(new);
          free(vlist);
          return 6;
        }
      }
      /* クリークが見つかった時の処理  〜 ここまで 〜 */
    } else {
      /* 再帰的処理 */
      if (newnum + 1 >= maxsize - degree){
        setBit(selected, old[i]);
        degree++;
        explore_numonly(new, newnum);
        resetBit(selected, old[i]);
        degree--;
      }
    }
  }

  free(num);
  free(new);

  return 0;
}

/*
  同じ大きさの二つの配列(dat, dat2)を
  第一配列の要素の降順に並べる
  要素数は最大でも500程度
  再帰呼出しのquick sortではなく敢えてshell sortにしてみる
*/
void sort_(int datnum, int *dat, int *dat2){
  int i, j, k;
  int h = 1;
  int t;

  i  = j  = k  = 0;

  while (h <= datnum) h = h * 3 + 1;
  h = h / 9;
  if (h < 1) h = 1;

  while (h > 0){
    for (i = h; i < datnum; i++){
      k = i;
      j = i - h;
      while (j >= 0 && dat[j] < dat[k]){
        t      = dat[j];
        dat[j] = dat[k];
        dat[k] = t;
        t      = dat2[j];
        dat2[j] = dat2[k];
        dat2[k] = t;
        j -= h;
        k -= h;
      }
    }
    h = h / 3;
  }
  return;
}

