/*
  ドッキンググラフのための関数
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "simcomp.h"

int make_DockingNode_Atom(SIMCOMP *, int, int);
int make_DockingNode_Atom2(SIMCOMP *, int, int);
int make_DockingNode_Bond(SIMCOMP *, int, int);
void make_DockingGraph_Atom(SIMCOMP *);
void make_DockingGraph_Bond(SIMCOMP *);
void check_effectiveNode_Atom(SIMCOMP *);
void check_effectiveNode_Bond(SIMCOMP *);

/*
  cpdmol_1 と cpdmol_2 のノード数をチェック
*/
int check_nudonum(SIMCOMP *simcomp){
  int min;
  double score;

  simcomp->status = 0;

  /* 特定の場合は何もせずに帰る */
  if (simcomp->cpdmol_1 == NULL || simcomp->cpdmol_2 == NULL){
    simcomp->status = 9;
  }

  /* ドッキングノードの作成方法に応じたノード数を設定 */
  switch(simcomp->node_mode){
  case _NT_ATOM_:
  case _NT_ATOM2_:
    simcomp->cpdmol_1->nodenum = simcomp->cpdmol_1->atomnum;
    simcomp->cpdmol_1->nodenum_sub = simcomp->cpdmol_1->atomnum_sub;
    simcomp->cpdmol_2->nodenum = simcomp->cpdmol_2->atomnum;
    simcomp->cpdmol_2->nodenum_sub = simcomp->cpdmol_2->atomnum_sub;
    break;
  case _NT_BOND_:
    simcomp->cpdmol_1->nodenum = simcomp->cpdmol_1->bondnum;
    simcomp->cpdmol_1->nodenum_sub = simcomp->cpdmol_1->bondnum_sub;
    simcomp->cpdmol_2->nodenum = simcomp->cpdmol_2->bondnum;
    simcomp->cpdmol_2->nodenum_sub = simcomp->cpdmol_2->bondnum_sub;
    break;
  }

  /* ノード数が 0 の時は、結果も自明である */
  if (simcomp->cpdmol_1->nodenum == 0 || simcomp->cpdmol_2->nodenum == 0){
    simcomp->status = 1;
  } else {
    min = (simcomp->cpdmol_1->nodenum_sub < simcomp->cpdmol_2->nodenum_sub) ?
           simcomp->cpdmol_1->nodenum_sub : simcomp->cpdmol_2->nodenum_sub;
    score = (double)min /
      (double)(simcomp->cpdmol_1->nodenum_sub + simcomp->cpdmol_2->nodenum_sub - min);

    if (simcomp->threshold > score) simcomp->status = 2;
  }

  return simcomp->status;
}

/* ドッキングノード領域を確保 */
void alloc_DockingNode(SIMCOMP *simcomp){
  int max;   /* ドッキングノード数の最大数 */

  switch(simcomp->node_mode){
  case _NT_ATOM_:
  case _NT_ATOM2_:
    /* ドッキングノード数は最大でも原子数をかけあわせたもの */
    max = simcomp->cpdmol_1->atomnum_sub * simcomp->cpdmol_2->atomnum_sub;

    simcomp->dnode = (DNODE *)malloc(sizeof(DNODE) * max);
    if (simcomp->dnode == NULL){
      simcomp->status = 6;
      return;
    }
    memset(simcomp->dnode, 0, sizeof(DNODE) * max);
    break;
  case _NT_BOND_:
    /* ドッキングノード数は最大でも結合数をかけあわせたもの */
    max = simcomp->cpdmol_1->bondnum_sub * simcomp->cpdmol_2->bondnum_sub;

    simcomp->dnode = (DNODE *)malloc(sizeof(DNODE) * max);
    if (simcomp->dnode == NULL){
      simcomp->status = 6;
      return;
    }
    memset(simcomp->dnode, 0, sizeof(DNODE) * max);
    break;
  default:  /* ノードの作成方法が不明 */
    simcomp->status = 2;
    break;
  }
  return;
}

/* ドッキングノードを作成 */
int make_DockingNode(SIMCOMP *simcomp, int atom_mode, int usedflag){
  int ret = 0;

  if (simcomp->status != 0) return ret;
  switch(simcomp->node_mode){
  case _NT_ATOM_:
    ret = make_DockingNode_Atom(simcomp, atom_mode, usedflag);
    break;
  case _NT_ATOM2_:
    ret = make_DockingNode_Atom2(simcomp, atom_mode, usedflag);
    break;
  case _NT_BOND_:
    ret = make_DockingNode_Bond(simcomp, atom_mode, usedflag);
    break;
  default:  /* ノードの作成方法が不明 */
    simcomp->status = 2;
    break;
  }

  return ret;
}

/* 原子によるドッキングノード作成 */
int make_DockingNode_Atom(SIMCOMP *simcomp, int atom_mode, int usedflag){
  int i, j, k;
  ATOM *q, *t;
  char *qid, *tid;
  double score;
  char ident;
  int nodenum = 0;
  int regist;

  qid = NULL;
  tid = NULL;

  /* 使用フラグをクリア(連続実行に備えて) */
  if (usedflag == _YES_){
    for (i = 0; i < simcomp->cpdmol_1->atomnum; i++)
      simcomp->cpdmol_1->atom[i].used = _NO_;
    for (i = 0; i < simcomp->cpdmol_2->atomnum; i++)
      simcomp->cpdmol_2->atom[i].used = _NO_;
  }

  for (i = 0; i < simcomp->cpdmol_1->atomnum; i++){
    q = &(simcomp->cpdmol_1->atom[i]);
    if (q->HAflag == _YES_) continue;

    for (j = 0; j < simcomp->cpdmol_2->atomnum; j++){
      t = &(simcomp->cpdmol_2->atom[j]);
      if (t->HAflag == _YES_) continue;

      /* Nodeの二重登録を防止 */
      if (usedflag == _NO_){
        regist = _NO_;
        for (k = 0; k < simcomp->nodenum; k++){
          if (simcomp->dnode[k].e1 == i && simcomp->dnode[k].e2 == j){
            regist = _YES_;
            break;
          }
        }
        if (regist == _YES_) continue;
      }

      switch(atom_mode){
      case _DC_NORMAL_:
        qid = q->nid;
        tid = t->nid;
        break;
      case _DC_CLASS_:
        qid = q->cid;
        tid = t->cid;
        break;
      case _DC_ATOM_:
        qid = q->aid;
        tid = t->aid;
        break;
      default:
        qid = q->nid;
        tid = t->nid;
        break;
      }

      score = 0.5;
      ident = 0;

      if (q->RAneighbor == _YES_ || t->RAneighbor == _YES_){
        if (strcmp(q->aid, t->aid) != 0){
          continue;
        }
      } else {
        if (strcmp(qid, tid) != 0){
          continue;
        } else {
          score = 1.0;
          ident = 1;
        }
      }
      simcomp->dnode[simcomp->nodenum + nodenum].e1 = i;
      simcomp->dnode[simcomp->nodenum + nodenum].e2 = j;
      simcomp->dnode[simcomp->nodenum + nodenum].score = score;
      simcomp->dnode[simcomp->nodenum + nodenum].ident = ident;
      nodenum++;

      if (usedflag == _YES_){
        simcomp->cpdmol_1->atom[i].used = _YES_;
        simcomp->cpdmol_2->atom[j].used = _YES_;
      }
    }
  }

  return nodenum;
}

/*
  同じ原子タイプの結合が相手分子の存在する場合、
  その結合に属する原子のみによるドッキングノード作成
*/
int make_DockingNode_Atom2(SIMCOMP *simcomp, int atom_mode, int usedflag){
  int i, j, k;
  int ok1, ok2;
  ATOM *q1, *q2, *t1, *t2;
  char *qid, *tid;
  double score;
  char ident;
  int nodenum = 0;
  int regist;

  qid = NULL;
  tid = NULL;

  /* 使用フラグをクリア(連続実行に備えて) */
  if (usedflag == _YES_){
    for (i = 0; i < simcomp->cpdmol_1->atomnum; i++)
      simcomp->cpdmol_1->atom[i].used = _NO_;
    for (i = 0; i < simcomp->cpdmol_2->atomnum; i++)
      simcomp->cpdmol_2->atom[i].used = _NO_;
  }
  for (i = 0; i < simcomp->cpdmol_1->bondnum; i++)
    simcomp->cpdmol_1->bond[i].used = _NO_;
  for (i = 0; i < simcomp->cpdmol_2->bondnum; i++)
    simcomp->cpdmol_2->bond[i].used = _NO_;

  /* 一旦結合の組み合わせでチェック */
  for (i = 0; i < simcomp->cpdmol_1->bondnum; i++){
    q1 = &(simcomp->cpdmol_1->atom[simcomp->cpdmol_1->bond[i].a1 - 1]);
    q2 = &(simcomp->cpdmol_1->atom[simcomp->cpdmol_1->bond[i].a2 - 1]);
    if (q1->HAflag == _YES_) continue;
    if (q2->HAflag == _YES_) continue;

    for (j = 0; j < simcomp->cpdmol_2->bondnum; j++){
      t1 = &(simcomp->cpdmol_2->atom[simcomp->cpdmol_2->bond[j].a1 - 1]);
      t2 = &(simcomp->cpdmol_2->atom[simcomp->cpdmol_2->bond[j].a2 - 1]);
      if (t1->HAflag == _YES_) continue;
      if (t2->HAflag == _YES_) continue;

      switch(atom_mode){
      case _DC_NORMAL_:
        qid = q1->nid;
        tid = t1->nid;
        break;
      case _DC_CLASS_:
        qid = q1->cid;
        tid = t1->cid;
        break;
      case _DC_ATOM_:
        qid = q1->aid;
        tid = t1->aid;
        break;
      default:
        qid = q1->nid;
        tid = t1->nid;
        break;
      }

      /* 結合をa1とa1、a2とa2で重ね合わせが可能か */
      ok1 = _NO_;
      ok2 = _NO_;
      if (q1->RAneighbor == _YES_ || t1->RAneighbor == _YES_){
        if (strcmp(q1->aid, t1->aid) == 0) ok1 = _YES_;
      } else {
        if (strcmp(qid, tid) == 0){
          ok1 = _YES_;
        }
      }
      if (q2->RAneighbor == _YES_ || t2->RAneighbor == _YES_){
        if (strcmp(q2->aid, t2->aid) == 0) ok2 = _YES_;
      } else {
        if (strcmp(qid, tid) == 0) ok2 = _YES_;
      }

      if (ok1 == _NO_ || ok2 == _NO_){
        /* 結合をa1とa2、a2とa1で重ね合わせが可能か */
        ok1 = _NO_;
        ok2 = _NO_;
        if (q1->RAneighbor == _YES_ || t2->RAneighbor == _YES_){
          if (strcmp(q1->aid, t2->aid) == 0) ok1 = _YES_;
        } else {
          if (strcmp(qid, tid) == 0){
            ok1 = _YES_;
          }
        }
        if (q2->RAneighbor == _YES_ || t1->RAneighbor == _YES_){
          if (strcmp(q2->aid, t1->aid) == 0) ok2 = _YES_;
        } else {
          if (strcmp(qid, tid) == 0) ok2 = _YES_;
        }
      }

      if (ok1 == _NO_ || ok2 == _NO_) continue;

      simcomp->cpdmol_1->bond[i].used = _YES_;
      simcomp->cpdmol_2->bond[j].used = _YES_;
    }
  }

  for (i = 0; i < simcomp->cpdmol_1->atomnum; i++){
    q1 = &(simcomp->cpdmol_1->atom[i]);
    if (q1->HAflag == _YES_) continue;

    ok1 = _NO_;
    for (k = 0; k < q1->bondnum; k++){
      if (simcomp->cpdmol_1->bond[q1->bond[k]].used == _YES_){
        ok1 = _YES_;
        break;
      }
    }
    if (ok1 == _NO_) continue;

    for (j = 0; j < simcomp->cpdmol_2->atomnum; j++){
      t1 = &(simcomp->cpdmol_2->atom[j]);
      if (t1->HAflag == _YES_) continue;

      ok2 = _NO_;
      for (k = 0; k < t1->bondnum; k++){
        if (simcomp->cpdmol_2->bond[t1->bond[k]].used == _YES_){
          ok2 = _YES_;
          break;
        }
      }
      if (ok2 == _NO_) continue;

      /* Nodeの二重登録を防止 */
      if (usedflag == _NO_){
        regist = _NO_;
        for (k = 0; k < simcomp->nodenum; k++){
          if (simcomp->dnode[k].e1 == i && simcomp->dnode[k].e2 == j){
            regist = _YES_;
            break;
          }
        }
        if (regist == _YES_) continue;
      }

      switch(atom_mode){
      case _DC_NORMAL_:
        qid = q1->nid;
        tid = t1->nid;
        break;
      case _DC_CLASS_:
        qid = q1->cid;
        tid = t1->cid;
        break;
      case _DC_ATOM_:
        qid = q1->aid;
        tid = t1->aid;
        break;
      default:
        qid = q1->nid;
        tid = t1->nid;
        break;
      }

      score = 0.5;
      ident = 0;

      if (q1->RAneighbor == _YES_ || t1->RAneighbor == _YES_){
        if (strcmp(q1->aid, t1->aid) != 0){
          continue;
        }
      } else {
        if (strcmp(qid, tid) != 0){
          continue;
        } else {
          score = 1.0;
          ident = 1;
        }
      }

      simcomp->dnode[simcomp->nodenum + nodenum].e1 = i;
      simcomp->dnode[simcomp->nodenum + nodenum].e2 = j;
      simcomp->dnode[simcomp->nodenum + nodenum].score = score;
      simcomp->dnode[simcomp->nodenum + nodenum].ident = ident;
      nodenum++;

      if (usedflag == _YES_){
        simcomp->cpdmol_1->atom[i].used = _YES_;
        simcomp->cpdmol_2->atom[j].used = _YES_;
      }
    }
  }

  return nodenum;
}

/* 結合によるドッキングノード作成 */
int make_DockingNode_Bond(SIMCOMP *simcomp, int atom_mode, int usedflag){
  int i, j, k;
  int ok1, ok2;
  ATOM *q1, *q2, *t1, *t2;
  char *qid1, *qid2, *tid1, *tid2;
  double score;
  char ident;
  int nodenum = 0;
  int regist;

  qid1 = NULL;
  qid2 = NULL;
  tid1 = NULL;
  tid2 = NULL;

  /* 使用フラグをクリア(連続実行に備えて) */
  if (usedflag == _YES_){
    for (i = 0; i < simcomp->cpdmol_1->bondnum; i++)
      simcomp->cpdmol_1->bond[i].used = _NO_;
    for (i = 0; i < simcomp->cpdmol_2->bondnum; i++)
      simcomp->cpdmol_2->bond[i].used = _NO_;
  }

  for (i = 0; i < simcomp->cpdmol_1->bondnum; i++){
    q1 = &(simcomp->cpdmol_1->atom[simcomp->cpdmol_1->bond[i].a1 - 1]);
    q2 = &(simcomp->cpdmol_1->atom[simcomp->cpdmol_1->bond[i].a2 - 1]);
    if (q1->HAflag == _YES_) continue;
    if (q2->HAflag == _YES_) continue;

    for (j = 0; j < simcomp->cpdmol_2->bondnum; j++){
      t1 = &(simcomp->cpdmol_2->atom[simcomp->cpdmol_2->bond[j].a1 - 1]);
      t2 = &(simcomp->cpdmol_2->atom[simcomp->cpdmol_2->bond[j].a2 - 1]);
      if (t1->HAflag == _YES_) continue;
      if (t2->HAflag == _YES_) continue;

      /* Nodeの二重登録を防止 */
      if (usedflag == _NO_){
        regist = _NO_;
        for (k = 0; k < simcomp->nodenum; k++){
          if (simcomp->dnode[k].e1 == i && simcomp->dnode[k].e2 == j){
            regist = _YES_;
            break;
          }
        }
        if (regist == _YES_) continue;
      }

      switch(atom_mode){
      case _DC_NORMAL_:
        qid1 = q1->nid;
        qid2 = q2->nid;
        tid1 = t1->nid;
        tid2 = t2->nid;
        break;
      case _DC_CLASS_:
        qid1 = q1->cid;
        qid2 = q2->cid;
        tid1 = t1->cid;
        tid2 = t2->cid;
        break;
      case _DC_ATOM_:
        qid1 = q1->aid;
        qid2 = q2->aid;
        tid1 = t1->aid;
        tid2 = t2->aid;
        break;
      default:
        qid1 = q1->nid;
        qid2 = q2->nid;
        tid1 = t1->nid;
        tid2 = t2->nid;
        break;
      }

      score = 1.0;
      ident = 1;

      /* 結合をa1とa1、a2とa2で重ね合わせが可能か */
      ok1 = _NO_;
      ok2 = _NO_;
      if (q1->RAneighbor == _YES_ || t1->RAneighbor == _YES_){
        if (strcmp(q1->aid, t1->aid) == 0){
          ok1 = _YES_;
          score = 0.5;
          ident = 0;
        }
      } else {
        if (strcmp(qid1, tid1) == 0){
          ok1 = _YES_;
        }
      }
      if (q2->RAneighbor == _YES_ || t2->RAneighbor == _YES_){
        if (strcmp(q2->aid, t2->aid) == 0){
          ok2 = _YES_;
          score = 0.5;
          ident = 0;
        }
      } else {
        if (strcmp(qid2, tid2) == 0) ok2 = _YES_;
      }

      if (ok1 == _NO_ || ok2 == _NO_){
        /* 結合をa1とa2、a2とa1で重ね合わせが可能か */
        ok1 = _NO_;
        ok2 = _NO_;
        if (q1->RAneighbor == _YES_ || t2->RAneighbor == _YES_){
          if (strcmp(q1->aid, t2->aid) == 0){
            ok1 = _YES_;
            score = 0.5;
            ident = 0;
          }
        } else {
          if (strcmp(qid1, tid2) == 0){
            ok1 = _YES_;
          }
        }
        if (q2->RAneighbor == _YES_ || t1->RAneighbor == _YES_){
          if (strcmp(q2->aid, t1->aid) == 0){
            ok2 = _YES_;
            score = 0.5;
            ident = 0;
          }
        } else {
          if (strcmp(qid2, tid1) == 0) ok2 = _YES_;
        }
      }

      if (ok1 == _NO_ || ok2 == _NO_) continue;

      simcomp->dnode[simcomp->nodenum + nodenum].e1 = i;
      simcomp->dnode[simcomp->nodenum + nodenum].e2 = j;
      simcomp->dnode[simcomp->nodenum + nodenum].score = score;
      simcomp->dnode[simcomp->nodenum + nodenum].ident = ident;
      nodenum++;

      if (usedflag == _YES_){
        simcomp->cpdmol_1->bond[i].used = _YES_;
        simcomp->cpdmol_2->bond[j].used = _YES_;
      }
    }
  }

  return nodenum;
}

/* 有効ノード数をチェック */
int check_effectiveNode(SIMCOMP *simcomp){
  switch(simcomp->node_mode){
  case _NT_ATOM_:
  case _NT_ATOM2_:
    check_effectiveNode_Atom(simcomp);
    break;
  case _NT_BOND_:
    check_effectiveNode_Bond(simcomp);
    break;
  default:  /* ノードの作成方法が不明 */
    simcomp->status = 2;
    break;
  }

  return simcomp->status;
}

void check_effectiveNode_Atom(SIMCOMP *simcomp){
  int i;
  int qnum = 0;
  int tnum = 0;
  int min;
  double score;

  for (i = 0; i < simcomp->cpdmol_1->atomnum; i++){
    if (simcomp->cpdmol_1->atom[i].used == _YES_) qnum++;
  }
  for (i = 0; i < simcomp->cpdmol_2->atomnum; i++){
    if (simcomp->cpdmol_2->atom[i].used == _YES_) tnum++;
  }

  min = (qnum < tnum) ? qnum : tnum;
  score = (double)min /
    (double)(simcomp->cpdmol_1->nodenum_sub + simcomp->cpdmol_2->nodenum_sub - min);
  if (simcomp->threshold > score) simcomp->status = 2;

  return;
}

void check_effectiveNode_Bond(SIMCOMP *simcomp){
  int i;
  int qnum = 0;
  int tnum = 0;
  int min;
  double score;

  for (i = 0; i < simcomp->cpdmol_1->bondnum; i++){
    if (simcomp->cpdmol_1->bond[i].used == _YES_) qnum++;
  }
  for (i = 0; i < simcomp->cpdmol_2->bondnum; i++){
    if (simcomp->cpdmol_2->bond[i].used == _YES_) tnum++;
  }

  min = (qnum < tnum) ? qnum : tnum;
  score = (double)min /
    (double)(simcomp->cpdmol_1->nodenum_sub + simcomp->cpdmol_2->nodenum_sub - min);
  if (simcomp->threshold > score) simcomp->status = 2;

  return;
}

/*
  ドッキンググラフを作る

  引数
      SIMCOMP *simcomp  ... SIMCOMPデータ

  戻り値

      結合マトリックスについて
      ・nodenum * columnnum の大きさの unsigned int 型配列
      ・要素は 0 または 1 で、対角成分は 0
             0 ... 結線無し
             1 ... 結線あり（両方ともつながっているか両方ともつながっていない）

      SCCS作成用 行列について
      ・nodenum * columnnum の大きさの unsigned int 型配列
      ・要素は 0 または 1 で、対角成分は 1
             0 ... 結線無し
             1 ... 結線あり（両方ともつながっている）
*/
void make_DockingGraph(SIMCOMP *simcomp){
  int i;

  simcomp->columnnum = (simcomp->nodenum + simcomp->nodetrick_num + _BIT_LENGTH_ - 1) / _BIT_LENGTH_;

  simcomp->matrix = ctl_mallocArray_uint2(simcomp->nodenum + simcomp->nodetrick_num, simcomp->columnnum + simcomp->nodetrick_num);
  simcomp->connect = ctl_mallocArray_uint2(simcomp->nodenum + simcomp->nodetrick_num, simcomp->columnnum + simcomp->nodetrick_num);

  if (simcomp->matrix == NULL || simcomp->connect == NULL){
    simcomp->status = 6;
    return;
  }

  /* SCCS作成用 行列の対角 */
  for (i = 0; i < simcomp->nodenum; i++){
    setBitInMatrix(simcomp->connect, i, i);
  }

  switch(simcomp->node_mode){
  case _NT_ATOM_:
  case _NT_ATOM2_:
    make_DockingGraph_Atom(simcomp);
    break;
  case _NT_BOND_:
    make_DockingGraph_Bond(simcomp);
    break;
  default:  /* ノードの作成方法が不明　(そもそもここにはこないはず) */
    simcomp->status = 2;
    break;
  }

  return;
}

void make_DockingGraph_Atom(SIMCOMP *simcomp){
  CPDMOL *mol1 = simcomp->cpdmol_1;
  CPDMOL *mol2 = simcomp->cpdmol_2;
  int i, j;
  int q, t;

  for (i = 0; i < simcomp->nodenum + simcomp->nodetrick_num - 1; i++){
    for (j = i + 1; j < simcomp->nodenum + simcomp->nodetrick_num; j++){
      if (simcomp->dnode[i].e1 == simcomp->dnode[j].e1) continue;
      if (simcomp->dnode[i].e2 == simcomp->dnode[j].e2) continue;

      /* 二つの原子の結合状態を求める */
      q = isAtomConnect(mol1, simcomp->dnode[i].e1, simcomp->dnode[j].e1);
      t = isAtomConnect(mol2, simcomp->dnode[i].e2, simcomp->dnode[j].e2);

      /* 結合状態が不一致なら次 */
      if (q != t) continue;

      setBitInMatrix(simcomp->matrix, i, j);
      setBitInMatrix(simcomp->matrix, j, i);
      if (q == _YES_){  /* 結合しているなら */
        setBitInMatrix(simcomp->connect, i, j);
        setBitInMatrix(simcomp->connect, j, i);
      }
    }
  }

  return;
}

void make_DockingGraph_Bond(SIMCOMP *simcomp){
  CPDMOL *mol1 = simcomp->cpdmol_1;
  CPDMOL *mol2 = simcomp->cpdmol_2;
  int i, j;
  int q, t;
  int ok;

  for (i = 0; i < simcomp->nodenum + simcomp->nodetrick_num - 1; i++){
    for (j = i + 1; j < simcomp->nodenum + simcomp->nodetrick_num; j++){
      ok = _NO_;
      if (simcomp->dnode[i].e1 == simcomp->dnode[j].e1) continue;
      if (simcomp->dnode[i].e2 == simcomp->dnode[j].e2) continue;

      /* 二つの結合で共通の原子を求める */
      q = commonAtom(&(mol1->bond[simcomp->dnode[i].e1]), 
                     &(mol1->bond[simcomp->dnode[j].e1]));
      t = commonAtom(&(mol2->bond[simcomp->dnode[i].e2]), 
                     &(mol2->bond[simcomp->dnode[j].e2]));

      if (q == -1 && t == -1){ /* 共通の原子ともに無し(ともに繋がっていない) */
        ok = _YES_;
      } else
      if (q == -1 || t == -1){ /* 共通の原子一方にあり */
        continue;
      }

      /* 共通の原子ともにあり(ともに繋がっている)　なので原子のタイプをチェック */
      /* comment out 20090819 */
/*
      if (ok == _NO_){
        if (mol1->atom[q].RAneighbor == _YES_ || mol2->atom[t].RAneighbor == _YES_){
          if (strcmp(mol1->atom[q].aid, mol2->atom[t].aid) == 0) ok = _YES_;
        } else {
          if (strcmp(mol1->atom[q].nid, mol2->atom[t].nid) == 0) ok = _YES_;
        }
      }
*/
      /* comment out end */

      setBitInMatrix(simcomp->matrix, i, j);
      setBitInMatrix(simcomp->matrix, j, i);
      if (q >= 0){
        setBitInMatrix(simcomp->connect, i, j);
        setBitInMatrix(simcomp->connect, j, i);
      }
    }
  }
  return;
}

/*
 2つの原子が結合しているか
*/
int isAtomConnect(CPDMOL *mol, int a1, int a2){
  int i;

  for (i = 0; i < mol->atom[a1].bondnum; i++){
    if (hasAtom(&(mol->bond[mol->atom[a1].bond[i]]), a2 + 1) == _YES_)
      return _YES_;
  }

  return _NO_;
}

/*
 2つの結合に共通する原子を返す
 無ければ-1を返す
*/
int commonAtom(BOND *b1, BOND *b2){
  int common = -1;

  if (b1 == NULL || b2 == NULL) return common;

  if (hasAtom(b1, b2->a1) == _YES_){
    common = b2->a1;
  } else
  if (hasAtom(b1, b2->a2) == _YES_){
    common = b2->a2;
  }

  return common;
}

/*
 結合が原子を含んできるか
 atomは1から数える
*/
int hasAtom(BOND *bond, int atom){
  int ret = _NO_;

  if (bond->a1 == atom || bond->a2 == atom) ret = _YES_;

  return ret;
}
