/*
  化合物グラフ
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "simcomp.h"
#include "classtable.h"

/* private function */
void setChiralPos(CPDMOL *);     /* キラルチェックに備えて、対象に成り得る部分を選別 */

CPDMOL *read_CPDMOL_KCF(FILE *fp, int chiral_check)
{
  CPDMOL *ret;
  char buff[BUFSIZ];
  char s1[BUFSIZ];
  char s2[BUFSIZ];
  char s3[BUFSIZ];
  char s4[BUFSIZ];
  int dummy;
  int rcode;
  int i;
  int j;
  int idx1;
  int idx2;

  /* ENTRY行を読み込む */
  if (fgets(buff, sizeof(buff), fp) == NULL) return NULL;
  if (strncmp(buff, "ENTRY", 5) != 0) return NULL;

  ret = (CPDMOL *)malloc(sizeof(CPDMOL));
  memset(ret, 0, sizeof(CPDMOL));

  if ((rcode = sscanf(buff, "%s %s", s1, s2)) != 2) return ret;
  ret->name = (char *)malloc(sizeof(char) * (strlen(s2) + 1));
  if (ret->name == NULL) return ret;
  memset(ret->name, 0, sizeof(char) * (strlen(s2) + 1));
  strcpy(ret->name, s2);

  /* NAME行を読み込む */
  if(fgets(buff, sizeof(buff), fp) == NULL) return ret;
  if(strncmp(buff, "NAME", 4) == 0){
    if (strlen(buff) > 12){
      sscanf(&buff[12], "%s", s1);
      if (strlen(s1) > 0){
        ret->definition = (char *)malloc(sizeof(char) * (strlen(s1) + 1));
        if (ret->definition == NULL) return ret;
        memset(ret->definition, 0, sizeof(char) * (strlen(s1) + 1));
        strcpy(ret->definition, s1);
      }
    }
  } else {
    ret->definition = NULL;
  }

  /* 原子情報への頭出し */
  while(strncmp(buff, "ATOM", 4) != 0 && strncmp(buff, "///", 3) != 0)
    if (fgets(buff, sizeof(buff), fp) == NULL) break;
  if (strncmp(buff, "///", 3) == 0) return ret;

  /* 原子数 */
  sscanf(buff, "%s %d", s1, &(ret->atomnum));
  if (ret->atomnum > 0){
    ret->atom = (ATOM *)malloc(sizeof(ATOM) * ret->atomnum);
    if (ret->atom == NULL){
      ret->atomnum = 0;  /* これを0にすれば、以降の処理がされなくなる */
      return ret;
    }
    memset(ret->atom, 0, sizeof(ATOM) * ret->atomnum);
  }

  /* 原子情報を読み込む */
  for (i = 0; i < ret->atomnum; i++){
    if (fgets(buff, sizeof(buff), fp) == NULL){
      ret->atomnum = 0;  /* これを0にすれば、以降の処理がされなくなる */
      return ret;
    }
    if(strncmp(buff, "///", 3) == 0){
      ret->atomnum = 0;  /* これを0にすれば、以降の処理がされなくなる */
      return ret;
    }
    if(buff[0] != ' ' && buff[0] != '\t' && buff[0] != '\n') break;

    s1[0] = '\0';
    s2[0] = '\0';
    s3[0] = '\0';
    s4[0] = '\0';
    if ((rcode = sscanf(buff, "%d %s %s %lf %lf %s %s", &dummy, s1, s2, &(ret->atom[i].x), &(ret->atom[i].y), s3, s4)) < 5){
      ret->atomnum = 0;  /* これを0にすれば、以降の処理がされなくなる */
      return ret;
    }
    ret->atom[i].nid = (char *)malloc(sizeof(char) * (strlen(s1) + 1));
    ret->atom[i].aid = (char *)malloc(sizeof(char) * (strlen(s2) + 1));
    if (ret->atom[i].nid == NULL || ret->atom[i].aid == NULL){
      ret->atomnum = 0;  /* これを0にすれば、以降の処理がされなくなる */
      return ret;
    }
    memset(ret->atom[i].nid, 0, sizeof(char) * (strlen(s1) + 1));
    strcpy(ret->atom[i].nid, s1);
    memset(ret->atom[i].aid, 0, sizeof(char) * (strlen(s2) + 1));
    strcpy(ret->atom[i].aid, s2);

    j = 0;
    while(strlen(class_table[j][0]) > 0){
      if (strcmp(ret->atom[i].nid, class_table[j][0]) == 0){
        ret->atom[i].cid = (char *)malloc(sizeof(char) * (strlen(class_table[j][1]) + 1));
        if (ret->atom[i].cid == NULL){
          ret->atomnum = 0;  /* これを0にすれば、以降の処理がされなくなる */
          return ret;
        }
        memset(ret->atom[i].cid, 0, sizeof(char) * (strlen(class_table[j][1]) + 1));
        strcpy(ret->atom[i].cid, class_table[j][1]);
        break;
      }
      j++;
    }
    if (ret->atom[i].cid == NULL){
      ret->atom[i].cid = (char *)malloc(sizeof(char) * (strlen(ret->atom[i].nid) + 1));
      if (ret->atom[i].cid == NULL){
        ret->atomnum = 0;  /* これを0にすれば、以降の処理がされなくなる */
        return ret;
      }
      memset(ret->atom[i].cid, 0, sizeof(char) * (strlen(ret->atom[i].nid) + 1));
      strcpy(ret->atom[i].cid, ret->atom[i].nid);
    }

    if (rcode > 5 && strlen(s3) > 0){
      ret->atom[i].prop1 = (char *)malloc(sizeof(char) * (strlen(s3) + 1));
      if (ret->atom[i].prop1 == NULL){
        ret->atomnum = 0;  /* これを0にすれば、以降の処理がされなくなる */
        return ret;
      }
      memset(ret->atom[i].prop1, 0, sizeof(char) * (strlen(s3) + 1));
      strcpy(ret->atom[i].prop1, s3);
    }
    if (rcode > 6 && strlen(s4) > 0){
      ret->atom[i].prop2 = (char *)malloc(sizeof(char) * (strlen(s4) + 1));
      if (ret->atom[i].prop2 == NULL){
        ret->atomnum = 0;  /* これを0にすれば、以降の処理がされなくなる */
        return ret;
      }
      memset(ret->atom[i].prop2, 0, sizeof(char) * (strlen(s4) + 1));
      strcpy(ret->atom[i].prop2, s4);
    }

    if(strcmp(ret->atom[i].nid, "H") == 0 ||
       strcmp(ret->atom[i].nid, "*") == 0){
      ret->atom[i].HAflag = 1;
    } else {
      ret->atom[i].HAflag = 0;
      ret->atomnum_sub++;
    }
  }

  /* 結合情報への頭出し */
  while(strncmp(buff, "BOND", 4) != 0 && strncmp(buff, "///", 3) != 0)
    if (fgets(buff, sizeof(buff), fp) == NULL) break;
  if(strncmp(buff, "///", 3) == 0) return ret;

  /* 結合数 */
  sscanf(buff, "%s %d", s1, &(ret->bondnum));
  if (ret->bondnum > 0){
    ret->bond = (BOND *)malloc(sizeof(BOND) * ret->bondnum);
    if (ret->bond == NULL){
      ret->atomnum = 0;  /* これを0にすれば、以降の処理がされなくなる */
      return ret;
    }
    memset(ret->bond, 0, sizeof(BOND) * ret->bondnum);
  }

  /* 結合情報を読み込む */
  for (i = 0; i < ret->bondnum; i++){
    if (fgets(buff, sizeof(buff), fp) == NULL){
      ret->atomnum = 0;  /* これを0にすれば、以降の処理がされなくなる */
      return ret;
    }
    if(strncmp(buff, "///", 3) == 0){
      ret->atomnum = 0;  /* これを0にすれば、以降の処理がされなくなる */
      return ret;
    }
    if(buff[0] != ' ' && buff[0] != '\t' && buff[0] != '\n') break;

    s1[0] = '\0';
    if ((rcode = sscanf(buff, "%d %d %d %d %s",
        &dummy, &(ret->bond[i].a1), &(ret->bond[i].a2),
        &(ret->bond[i].order), s1)) < 4) continue;

    if (rcode > 4 && strlen(s1) > 0){
      ret->bond[i].prop = (char *)malloc(sizeof(char) * (strlen(s1) + 1));
      if (ret->bond[i].prop == NULL){
        ret->atomnum = 0;  /* これを0にすれば、以降の処理がされなくなる */
        return ret;
      }
      memset(ret->bond[i].prop, 0, sizeof(char) * (strlen(s1) + 1));
      strcpy(ret->bond[i].prop, s1);
    }
    if ((ret->bond[i].a1 > 0 && ret->bond[i].a1 <= ret->atomnum) &&
        (ret->bond[i].a2 > 0 && ret->bond[i].a2 <= ret->atomnum)){
      ret->atom[ret->bond[i].a1 - 1].bondnum++;
      ret->atom[ret->bond[i].a2 - 1].bondnum++;
      if (ret->atom[ret->bond[i].a1 - 1].HAflag == _NO_ &&
          ret->atom[ret->bond[i].a2 - 1].HAflag == _NO_)
        ret->bondnum_sub++;
    }
  }

  /* デリミタまで進む */
  while(strncmp(buff, "///", 3) != 0) 
    if (fgets(buff, sizeof(buff), fp) == NULL) break;

  /* ATOM構造体に、原子が結合情報に関連している数分の領域を確保 */
  for (i = 0; i < ret->atomnum; i++){
    ret->atom[i].bond = (int *)malloc(sizeof(int) * ret->atom[i].bondnum);
    if (ret->atom[i].bond == NULL){
      ret->atomnum = 0;  /* これを0にすれば、以降の処理がされなくなる */
      return ret;
    }
    memset(ret->atom[i].bond, -1, sizeof(int) * ret->atom[i].bondnum);
  }

  /* 分子データ内の結合配列の番号(先頭INDEX 0)を設定 */
  for (i = 0; i < ret->bondnum; i++){
    idx1 = ret->bond[i].a1 - 1;
    idx2 = ret->bond[i].a2 - 1;

    for (j = 0; j < ret->atom[idx1].bondnum; j++){
      if (ret->atom[idx1].bond[j] == -1){
         ret->atom[idx1].bond[j] = i;
         break;
      }
    }
    for (j = 0; j < ret->atom[idx2].bondnum; j++){
      if (ret->atom[idx2].bond[j] == -1){
         ret->atom[idx2].bond[j] = i;
         break;
      }
    }
  }

  /* Rの近傍チェック */
  for (i = 0; i < ret->atomnum; i++){
    if (strcmp(ret->atom[i].nid, "R") == 0 ||
        strcmp(ret->atom[i].nid, "*") == 0){
      for (j = 0; j < ret->atom[i].bondnum; j++){
        idx1 = ret->bond[ret->atom[i].bond[j]].a1 - 1;
        if (idx1 == i) idx1 = ret->bond[ret->atom[i].bond[j]].a2 - 1;
        ret->atom[idx1].RAneighbor = 1;
      }
    }
  }

  /* キラルチェックに備えて、対象に成り得る部分を選別 */
  if (chiral_check != _CC_NONE_) setChiralPos(ret);

  return ret;
}

/* キラルチェックに備えて、対象に成り得る部分を選別 */
void setChiralPos(CPDMOL *mol){
  int i;
  int a1, a2;

  /* 明示された水素数初期化 */
  for (i = 0; i < mol->atomnum; i++)
    mol->atom[i].nH = 0;

  for (i = 0; i < mol->bondnum; i++){
    a1 = mol->bond[i].a1 - 1;
    a2 = mol->bond[i].a2 - 1;

    /* 明示された水素数 */
    if (strcmp(mol->atom[a1].aid, "H") == 0) mol->atom[a2].nH++;
    if (strcmp(mol->atom[a2].aid, "H") == 0) mol->atom[a1].nH++;

    /* trans/cis */
    if (mol->bond[i].order == 2){
      /* 二重結合以外の結合が1つか2つでないとcis/transの区別は成立しない */
      if (mol->atom[a1].bondnum != 2 && mol->atom[a1].bondnum != 3) continue;
      if (mol->atom[a2].bondnum != 2 && mol->atom[a2].bondnum != 3) continue;
      /* 芳香環(C8、N5)原子同士の二重結合には適用しない */
      /* 二重結合であるなら芳香環の一部のはず */
      if ((strncmp(mol->atom[a1].nid, "C8", 2) == 0 ||
           strncmp(mol->atom[a1].nid, "N5", 2) == 0) &&
          (strncmp(mol->atom[a2].nid, "C8", 2) == 0 ||
           strncmp(mol->atom[a2].nid, "N5", 2) == 0)) continue;
      mol->bond[i].geometry = _YES_;
    }
  }

  /* 立体の結合の有無をチェック */
  for (i = 0; i < mol->bondnum; i++){
    if (mol->bond[i].prop == NULL) continue;
    if (strcasecmp(mol->bond[i].prop, "#UP") == 0 ||
        strcasecmp(mol->bond[i].prop, "#DOWN") == 0){
      mol->atom[mol->bond[i].a1 - 1].chiral = _YES_;
      mol->atom[mol->bond[i].a2 - 1].chiral = _YES_;
    }
  }

  /*
     結合、水素の数チェック
         対象元素　　　　C、N、S、P
         不斉となるのは　結合数=4　水素数1以下
                         結合数=3　水素数0(Nの場合、対象外とする)
   */
  for (i = 0; i < mol->atomnum; i++){
    if (mol->atom[i].chiral == _NO_) continue;

    if (strcmp(mol->atom[i].aid, "C") == 0 ||
        strcmp(mol->atom[i].aid, "N") == 0 ||
        strcmp(mol->atom[i].aid, "P") == 0 ||
        strcmp(mol->atom[i].aid, "S") == 0){
      if (mol->atom[i].bondnum == 3){
        if (mol->atom[i].nH > 0)
          mol->atom[i].chiral = _NO_;
        if (strcmp(mol->atom[i].aid, "N") == 0)
          mol->atom[i].chiral = _NO_;
      } else
      if (mol->atom[i].bondnum == 4){
        if (mol->atom[i].nH > 1)
          mol->atom[i].chiral = _NO_;
      } else {
        mol->atom[i].chiral = _NO_;
      }
    }
  }
  return;
}

/* CPDMOL内の領域開放 */
void clear_CPDMOL(CPDMOL *cpdmol){
  int i;

  if (cpdmol == NULL) return;

  if (cpdmol->name != NULL){
    free(cpdmol->name);
  }
  if (cpdmol->definition != NULL){
    free(cpdmol->definition);
  }
  if (cpdmol->atom != NULL){
    for (i = 0; i < cpdmol->atomnum; i++){
      if (cpdmol->atom[i].nid != NULL){
        free(cpdmol->atom[i].nid);
      }
      if (cpdmol->atom[i].cid != NULL){
        free(cpdmol->atom[i].cid);
      }
      if (cpdmol->atom[i].aid != NULL){
        free(cpdmol->atom[i].aid);
      }
      if (cpdmol->atom[i].prop1 != NULL){
        free(cpdmol->atom[i].prop1);
      }
      if (cpdmol->atom[i].prop2 != NULL){
        free(cpdmol->atom[i].prop2);
      }
      if (cpdmol->atom[i].bond != NULL){
        free(cpdmol->atom[i].bond);
      }
    }
    free(cpdmol->atom);
  }
  if (cpdmol->bond != NULL){
    for (i = 0; i < cpdmol->bondnum; i++){
      if (cpdmol->bond[i].prop != NULL){
        free(cpdmol->bond[i].prop);
      }
    }
    free(cpdmol->bond);
  }

  free(cpdmol);

  return;
}

void freelist_CPDMOL(CPDMOL **mol){
  CPDMOL* pmol;

  pmol = *mol;

  while(pmol != NULL){
    CPDMOL* pNext = pmol->pNext;
    clear_CPDMOL(pmol);
    pmol = pNext;
  }

  *mol = NULL;
}

