/*
  exe1.c
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "simcomp.h"

/* 計算結果のステータス */
/*
  0 ... 正常終了
  1 ... cpdmol のいずれかのノード数が 0 であった
  2 ... ドッキングノード数が 0 であった
  3 ... ドッキング時にノードオーバーフローを起こした　←　もう発生しない
  4 ... クリークオーバーフローを起こした（内部エラー）　←　もう発生しない
  9 ... cpdmol の何れかが NULL であった（入力エラー）
*/

/* private function */
double expect_similarity_Atom(SIMCOMP*);
double expect_similarity_Bond(SIMCOMP*);

void compute_trick();
void extract_Node(SIMCOMP *);
void extract_Node_byAtom(SIMCOMP *);
void extract_Node_byBond(SIMCOMP *);
int extend_MCS(SIMCOMP *);
int extend_MCS_byAtom(SIMCOMP *);
int extend_MCS_byBond(SIMCOMP *);
void output_kcfstyle(FILE *, SIMCOMP *);
void output_kcfstyle_cpd(FILE *, char *, CPDMOL *);
void output_kcfstyle_cpd_withColor(FILE *, char *, CPDMOL *, CPDMOL *,
  SIMCOMP *);
void make_alignInfo(SIMCOMP *);

static const unsigned int bitflags[] = {    /* 先頭からのビット位置 */
  0x80000000, 0x40000000, 0x20000000, 0x10000000,
  0x08000000, 0x04000000, 0x02000000, 0x01000000,
  0x00800000, 0x00400000, 0x00200000, 0x00100000,
  0x00080000, 0x00040000, 0x00020000, 0x00010000,
  0x00008000, 0x00004000, 0x00002000, 0x00001000,
  0x00000800, 0x00000400, 0x00000200, 0x00000100,
  0x00000080, 0x00000040, 0x00000020, 0x00000010,
  0x00000008, 0x00000004, 0x00000002, 0x00000001,
  0x00000000
};

/*********************************************************************/

/* cpdmol_1 と cpdmol_2 の間で類似度を見積もる */
double assume_similarity(SIMCOMP *simcomp){
  int min;

  min = (simcomp->cpdmol_1->nodenum < simcomp->cpdmol_2->nodenum) ?
         simcomp->cpdmol_1->nodenum : simcomp->cpdmol_2->nodenum;

  return 1.0 * (double)min / (double)(simcomp->cpdmol_1->nodenum +
                                      simcomp->cpdmol_2->nodenum - min);
}

/* 有効ノード数より、類似度の最大値を見積もる */
double expect_similarity(SIMCOMP *simcomp){
  double ret = 0.0;

  switch(simcomp->node_mode){
  case _NT_ATOM_:
  case _NT_ATOM2_:
    ret = expect_similarity_Atom(simcomp);
    break;
  case _NT_BOND_:
    ret = expect_similarity_Bond(simcomp);
    break;
  default:  /* ノードの作成方法が不明　(そもそもここにはこないはず) */
    simcomp->status = 2;
    break;
  }
  return ret;
}

double expect_similarity_Atom(SIMCOMP *simcomp){
  int i;
  int min;
  int q = 0;
  int t = 0;

  for (i = 0 ; i < simcomp->cpdmol_1->atomnum; i++){
    if (simcomp->cpdmol_1->atom[i].used == _YES_) q++;
  }
  for (i = 0 ; i < simcomp->cpdmol_2->atomnum; i++){
    if (simcomp->cpdmol_2->atom[i].used == _YES_) t++;
  }

  min = (q < t) ? q : t;
  return 1.0 * (double)min / (double)(simcomp->cpdmol_1->nodenum +
                                      simcomp->cpdmol_2->nodenum - min);
}

double expect_similarity_Bond(SIMCOMP *simcomp){
  int i;
  int min;
  int q = 0;
  int t = 0;

  for (i = 0 ; i < simcomp->cpdmol_1->bondnum; i++){
    if (simcomp->cpdmol_1->bond[i].used == _YES_) q++;
  }
  for (i = 0 ; i < simcomp->cpdmol_2->bondnum; i++){
    if (simcomp->cpdmol_2->bond[i].used == _YES_) t++;
  }

  min = (q < t) ? q : t;
  return 1.0 * (double)min / (double)(simcomp->cpdmol_1->nodenum +
                                      simcomp->cpdmol_2->nodenum - min);
}

/*
  cpdmol_1 と cpdmol_2 の間で類似度を計算する

  status = 0 or 4 のときは tanimoto coeff. を返し、それ以外では 0.0 を返す
*/
double compute_similarity(SIMCOMP* simcomp)
{
  double old = 0.0;

  /* クリークファインディング */
  clifinder2(simcomp);

  /* 従来のタニモト係数 */
  old = 1.0 * simcomp->mcs_max / (simcomp->cpdmol_1->nodenum_sub +
                       simcomp->cpdmol_2->nodenum_sub - simcomp->mcs_max);

  /*
    MCS, SCCSを用意
  */
  score_MCS_list(simcomp);  /* この中でSCCSも構築 */

  /* 結果を適当に修正するためのトリック */
  if (simcomp->exec_mode & (_EX_TRICK0_ | _EX_TRICK2_)) compute_trick(simcomp);

  /* 最大スコアの MCS を設定 */
  set_maxMCS(simcomp);

  /* トリックを使うとスコアが変化する　この値はcompsearch出力の是非にかかわる */
  if (simcomp->exec_mode & (_EX_TRICK0_ | _EX_TRICK2_)){
    old = 1.0 * simcomp->mcs_max / (simcomp->cpdmol_1->nodenum_sub +
                         simcomp->cpdmol_2->nodenum_sub - simcomp->mcs_max);
  }

  return old;
}

/*
  計算結果を書き出す

  _OS_NORMAL_    ... %s %d %s %d %d %f %f (%f %f) [no] [co]
  _OS_KCFFORMAT_ ... ENTRY       A00000            ReactantPair
                     NAME        %s_%s
                     COMPOUND    %s  %s
                                 %s  %s
  _OS_KCFFORMAT_ ... ALIGN       %d
                                 %-3d %3d:%3s %3d:%3s %s

  _OS_KCFFORMAT_ ... ENTRY1(2)
                       COMPOUND  %s
                       ATOM      %-3d
                                 %-3d %3s %2s %3d:%3s %3d:%3s
                       BOND      %-3d
                                 %-3d %3d %3d %d
                     ///


  status は の実行時に設定される
     status = 0 ... 普通に出力
     status = 1 ... cpdmol のいずれかのノード数が 0
     status = 2 ... ドッキングノード数が 0
     status = 3 ... ノードオーバーフロー　←　もう発生しない
     status = 4 ... クリークオーバーフロー　←　もう発生しない
     status = 5 ... 最大クリークのスコアが閾値に届かない
     status = 6 ... メモリ確保に失敗
     status = 9 ... 出力しない
*/
void output_similarity(FILE *fp, SIMCOMP *simcomp)
{
  double old    = 0.0; /* いわゆるタニモト係数 */
  double score  = 0.0; /* スコア (total)     */
  double scorei = 0.0; /* スコア (identical) */
  int i;
  CPDMOL *cpdmol1;
  CPDMOL *cpdmol2;

  cpdmol1 = simcomp->cpdmol_1;
  cpdmol2 = simcomp->cpdmol_2;

  if (simcomp->output_mode & _OS_KCFFORMAT_){
    make_alignInfo(simcomp);
    output_kcfstyle(fp, simcomp);
    return;
  }
  /* add 20091021 start */
  else if(simcomp->output_mode & _OS_KCFFORMAT_COLOR2_){
    make_alignInfo(simcomp);
    output_kcfstyle(fp, simcomp);
    return;
  }
  else if(simcomp->output_mode & _OS_KCFFORMAT_COLOR_){
    make_alignInfo(simcomp);
    output_kcfstyle(fp, simcomp);
    return;
  }
  /* add end */

	/* add 20090727 start */
  if (simcomp->status == 2) return;
	/* add end */

  if (simcomp->status == 9) return;

/* modify 20090806 */
/*
  fprintf(fp, "%s\t%d\t%s\t%d\t",
          cpdmol1->name, cpdmol1->nodenum_sub,
          cpdmol2->name, cpdmol2->nodenum_sub);
*/
/* modify end */


  if (simcomp->status == 0 || simcomp->status == 4){
    old = 1.0 * (simcomp->mcs_max - simcomp->max_mcs->chiral_penalty
                -simcomp->max_mcs->atom_penalty) /
                (simcomp->cpdmol_1->nodenum_sub +
                 simcomp->cpdmol_2->nodenum_sub
                  - (simcomp->mcs_max - simcomp->max_mcs->chiral_penalty
                    - simcomp->max_mcs->atom_penalty));

    if (simcomp->max_mcs != NULL) score = simcomp->max_mcs->score;

    for (i = 0; i < simcomp->nodenum + simcomp->nodetrick_num; i++){
      if (getBit(simcomp->max_mcs->vlist, i) != 0 &&
          simcomp->dnode[i].ident != 0){
        scorei += simcomp->dnode[i].score;
      }
    }
  }

  /* add 20090806 */
  if(score < simcomp->threshold) return;

  fprintf(fp, "%s\t%d\t%s\t%d\t",
          cpdmol1->name, cpdmol1->nodenum_sub,
          cpdmol2->name, cpdmol2->nodenum_sub);
  /* add end */

  fprintf(fp, "%d\t%f", simcomp->mcs_max, old);
  fprintf(fp, "\t%f (%f %f)", score, scorei, score - scorei);
  if (simcomp->status == 3) fprintf(fp, " (no:%d)", simcomp->nodenum);
  fprintf(fp, "\n");

  return;
}

void output_kcfstyle(FILE *fp, SIMCOMP *simcomp)
{
  int i, cnt;
  int idx1, idx2;
  int chiral = _NO_;
  char *ptr1, *ptr2;
  CPDMOL *mol1;
  CPDMOL *mol2;
  /* add 20090806 */
	double score;
  /* add end */

  if (simcomp->status == 9) return;

  /* add 20090806 */
	else if(simcomp->status == 2) return;

  if (simcomp->max_mcs != NULL) score = simcomp->max_mcs->score;
	else score = 0.0;

	if(score < simcomp->threshold) return;
  /* add end */

  mol1 = simcomp->cpdmol_1;
  mol2 = simcomp->cpdmol_2;

  fprintf(fp, "ENTRY       RP00000                     ReactantPair\n");
  fprintf(fp, "NAME        %s_%s\n",
          (mol1->name != NULL) ? mol1->name : "(null)",
          (mol2->name != NULL) ? mol2->name : "(null)");
  fprintf(fp, "COMPOUND    %s  %-58.58s\n",
          (mol1->name != NULL) ? mol1->name : "C00000",
          (mol1->definition != NULL) ? mol1->definition : "");
  fprintf(fp, "            %s  %-58.58s\n",
          (mol2->name != NULL) ? mol2->name : "C00000",
          (mol2->definition != NULL) ? mol2->definition : "");

/* modify 20090806 */
/*
  if ((simcomp->status != 0 && simcomp->status != 4) || simcomp->mcs_max == 0){
    if(simcomp->output_mode & _OS_KCFFORMAT_){
      output_kcfstyle_cpd(fp, "ENTRY1", mol1);
      output_kcfstyle_cpd(fp, "ENTRY2", mol2);
    }
    else if(simcomp->output_mode & _OS_KCFFORMAT_COLOR_){
      output_kcfstyle_cpd_withColor(fp, "ENTRY1", mol1, mol1, simcomp);
      output_kcfstyle_cpd_withColor(fp, "ENTRY2", mol2, mol1, simcomp);
    }
    fprintf(fp, "///\n");
    return;
  }
*/
/* modify end */

  /* add 20090806 */
	if(score == 0.0){
  	fprintf(fp, "ALIGN       %d  #similarity=%f\n", 0, 0.0);
  	fprintf(fp, "///\n");

		return;
	}
	/* add end */

  fprintf(fp, "ALIGN       %d  #similarity=%f\n", simcomp->alignnum,
                 1.0 * (simcomp->mcs_max - simcomp->max_mcs->chiral_penalty
                  -simcomp->max_mcs->atom_penalty) /
                  (simcomp->cpdmol_1->nodenum_sub +
                   simcomp->cpdmol_2->nodenum_sub
                    - (simcomp->mcs_max - simcomp->max_mcs->chiral_penalty
                      -simcomp->max_mcs->atom_penalty)));

  /*
     原子、結合のマッピング作成をキラルチェックと共用する際、
     ALIGN出力を　SCCS順から、ナンバリング順に変更
   */
  cnt = 0;
  for (i = 0; i < simcomp->cpdmol_1->atomnum; i++){
    chiral = _NO_;
    if (simcomp->max_mcs->atommap[i] != 0){
      idx1 = i;                                  /* クエリーの原子 */
      idx2 = simcomp->max_mcs->atommap[i];       /* ターゲットの原子(1から数える) */
      if (idx2 < 0){
        idx2 = -idx2;
        chiral = _YES_;
      } else {
        chiral = _NO_;
      }

      ptr1 = simcomp->cpdmol_1->atom[idx1].nid;  /* クエリーの原子タイプ */
      ptr2 = simcomp->cpdmol_2->atom[idx2 - 1].nid;  /* ターゲットの原子タイプ */

      cnt++;
      fprintf(fp, "            %-3d %3d:%-3s %3d:%-3s",
              cnt, idx1 + 1, ptr1, idx2, ptr2);
      if (chiral == _YES_) fprintf(fp, " #R");   /* キラルチェックで不一致 */
      fprintf(fp, "\n");
    }
  }

/* add 20091023 start */
  if (simcomp->output_mode & _OS_KCFFORMAT_COLOR2_){
    fprintf(fp, "BALIGN\n");
    for(i = 0; i < mol1->bondnum; i++){
      if(simcomp->max_mcs->bondmap[i] != 0){
        fprintf(fp, "            %-3d %-3d\n", i+1, simcomp->max_mcs->bondmap[i]);
      }
    }
    fprintf(fp, "///\n");
    return;
  }
/* add end */

/* add 20091021 start */
  if (simcomp->match_mode != _MA_ENTRYinDB_ || 
      simcomp->output_mode & _OS_KCFFORMAT_COLOR_){
    /* add, modify 20091021 start */
    if(simcomp->output_mode & _OS_KCFFORMAT_){
      output_kcfstyle_cpd(fp, "ENTRY1", mol1);
      output_kcfstyle_cpd(fp, "ENTRY2", mol2);
    }
    else if(simcomp->output_mode & _OS_KCFFORMAT_COLOR_){
      output_kcfstyle_cpd_withColor(fp, "ENTRY1", mol1, mol1, simcomp);
      output_kcfstyle_cpd_withColor(fp, "ENTRY2", mol2, mol1, simcomp);
    }
    /* add, modify end */
  }
/* add end*/
  fprintf(fp, "///\n");
}

void output_kcfstyle_cpd(FILE *fp, char *tag, CPDMOL *mol)
{
  int i;

  fprintf(fp, "%s\n", tag);
  fprintf(fp, "  COMPOUND  %s\n",
          (mol->name != NULL) ? mol->name : "(null)");
  fprintf(fp, "  ATOM      %d\n", mol->atomnum);
  for(i = 0; i < mol->atomnum; i++){
    fprintf(fp, "            %-3d %-3s %-2s %9.4f %9.4f",
                 i + 1, mol->atom[i].nid, mol->atom[i].aid, mol->atom[i].x, mol->atom[i].y);
    if (mol->atom[i].prop1 != NULL && strlen(mol->atom[i].prop1) > 0) fprintf(fp, " %s", mol->atom[i].prop1);
    if (mol->atom[i].prop2 != NULL && strlen(mol->atom[i].prop2) > 0) fprintf(fp, " %s", mol->atom[i].prop2);
    fprintf(fp, "\n");
  }
  fprintf(fp, "  BOND      %d\n", mol->bondnum);
  for(i = 0; i < mol->bondnum; i++){
    fprintf(fp, "            %-3d %3d %3d %d",
                 i + 1, mol->bond[i].a1, mol->bond[i].a2, mol->bond[i].order);
    if (mol->bond[i].prop != NULL && strlen(mol->bond[i].prop) > 0) fprintf(fp, " %s", mol->bond[i].prop);

    fprintf(fp, "\n");
  }
}

/* add 20091021 start */
/* kcfファイルの結合情報に、色情報を付加したものを出力する */
void output_kcfstyle_cpd_withColor(FILE *fp, char *tag, 
  CPDMOL *mol, CPDMOL *mol2, SIMCOMP *simcomp)
{
  int i, j;
  short int flg = 1;

  if(mol == mol2){
    flg = 0;
  }

  fprintf(fp, "%s\n", tag);
  fprintf(fp, "  COMPOUND  %s\n",
          (mol->name != NULL) ? mol->name : "(null)");


  /* ATOM情報の出力 */
  fprintf(fp, "  ATOM      %d\n", mol->atomnum);
  for(i = 0; i < mol->atomnum; i++){
    fprintf(fp, "            %-3d %-3s %-2s %9.4f %9.4f",
                 i + 1, mol->atom[i].nid, mol->atom[i].aid, mol->atom[i].x, mol->atom[i].y);
    if (mol->atom[i].prop1 != NULL && strlen(mol->atom[i].prop1) > 0) fprintf(fp, " %s", mol->atom[i].prop1);
    if (mol->atom[i].prop2 != NULL && strlen(mol->atom[i].prop2) > 0) fprintf(fp, " %s", mol->atom[i].prop2);

    /* クエリ分子の色情報の出力 */
    if(flg == 0){
        if(simcomp->max_mcs->atommap[i] != 0){
          fprintf(fp , " #ff0000");
        }
    }
    /* ターゲット分子の色情報の出力 */
    else{
       for(j=0; j < mol2->atomnum; j++){
         if(i+1 == simcomp->max_mcs->atommap[j]){
           fprintf(fp , " #ff0000");
           break;
         } 
       }
    }

    fprintf(fp, "\n");
  }

  /* BOND情報の出力 */
  fprintf(fp, "  BOND      %d\n", mol->bondnum);
  for(i = 0; i < mol->bondnum; i++){
    fprintf(fp, "            %-3d %3d %3d %d",
                 i + 1, mol->bond[i].a1, mol->bond[i].a2, mol->bond[i].order);
    if (mol->bond[i].prop != NULL && strlen(mol->bond[i].prop) > 0) fprintf(fp, " %s", mol->bond[i].prop);

    /* クエリ分子の色情報の出力 */
    if(flg == 0){
      if(simcomp->max_mcs->bondmap[i] != 0){
        fprintf(fp , " #ff0000");
      } 
    }
    /* ターゲット分子の色情報の出力 */
    else{
       for(j=0; j < mol2->bondnum; j++){
         if(i+1 == simcomp->max_mcs->bondmap[j]){
           fprintf(fp , " #ff0000");
           break;
         } 
       }
    }



    fprintf(fp, "\n");
  }
}
/* add end */


/******************************************************************/

/*
  計算結果のつじつまを合わせるための幾つかのトリック

  ０・一つ回りの原子を見ていってそれがドッキングノードに含まれていたら、
      クリークのノードとして採用する、さらに SCCS の計算も行う

  １・サイズが小さい（1個）コンポーネント（SCCS）を、除去する
      または、最大サイズの SCCS 以外を取り除く

  ２・一つ回りの原子を見ていってそれがドッキングノードに含まれていたら、
      クリークのノードとして採用する

  ３・新たに取れるノードがなくなるまで、１と２を繰り返す

  ４・SCCS の計算をして終了

  対象になるのは、set_maxMCS() で設定される MCS

  これらの作業により変更されるのは、

  mcslist[(最大のもの)];  -> vnum, vlist, score が適切に変更される
  mcslist[(上記以外)];    -> inactive 化される
  mcs_max;                -> 上記の vnum と同じ値になる
  max_mcs_index;          -> 対象 MCS を指す

*/
void compute_trick(SIMCOMP *simcomp){
  int i, j;
  int sccsnum;
  int ret;

  if (simcomp->mcslist.num == 0) return;
  /* 最大スコアの MCS を設定 */
  set_maxMCS(simcomp);

  /* max 以外の MCS を inactive(vnum = 0) にする */
  set_inactive(simcomp);

  /* extend_MCSの前準備 */
  extract_Node(simcomp);

  /* 一回だけ max_MCS を伸ばす */
#ifdef DEBUG
if (simcomp->debug_mode % 2 == 1){
printf("call extend_MCS\n");
}
#endif
  ret = extend_MCS(simcomp);

  /* max_MCSのスコアを設定 */
  if (ret != 0) score_maxMCS(simcomp);
  ret = 0;

  /* モードによって動作が異なる */
  if(simcomp->exec_mode & _EX_TRICK0_){
    /* _EX_TRICK0_ : 最大の SCCS を一つだけ残す */
    if (simcomp->max_mcs->sccsnum > 1){
      for (i = 1; i < simcomp->max_mcs->sccsnum; i++){
        if (simcomp->max_mcs->sccs[i].vlist != NULL){
          free(simcomp->max_mcs->sccs[i].vlist);
          ret = 1;
        }
      }

      simcomp->max_mcs->sccsnum = 1;
      simcomp->max_mcs->vnum = simcomp->max_mcs->sccs[0].vnum;
      simcomp->max_mcs->score = simcomp->max_mcs->sccs[0].score;
      for (i = 0; i < simcomp->columnnum; i++)
        simcomp->max_mcs->vlist[i] = simcomp->max_mcs->sccs[0].vlist[i];
    }
  } else if(simcomp->exec_mode & _EX_TRICK2_){
    /* _EX_TRICK2_ : SCCS を除去せずに extend だけ行う */
  } else {
    return;
  }

  /* ノードの作成方法が原子の場合、サイズ１のSCCSを削除 */
  if (simcomp->node_mode == _NT_ATOM_ || simcomp->node_mode == _NT_ATOM2_ ){
    sccsnum = simcomp->max_mcs->sccsnum;
    simcomp->max_mcs->sccsnum = 0;
    simcomp->max_mcs->vnum = 0;
    simcomp->max_mcs->score = 0.0;
    for (i = 0; i < simcomp->columnnum; i++)
      simcomp->max_mcs->vlist[i] = 0;

    for (i = 0; i < sccsnum; i++){
      if (simcomp->max_mcs->sccs[i].vnum == 1){
        simcomp->max_mcs->sccs[i].vnum = 0;
        simcomp->max_mcs->sccs[i].score = 0.0;
        if (simcomp->max_mcs->sccs[i].vlist != NULL){
          free(simcomp->max_mcs->sccs[i].vlist);
          simcomp->max_mcs->sccs[i].vlist = NULL;
        }
        ret = 1;
        continue;
      }

      simcomp->max_mcs->sccsnum++;
      simcomp->max_mcs->vnum += simcomp->max_mcs->sccs[i].vnum;
      simcomp->max_mcs->score += simcomp->max_mcs->sccs[i].score;
      for (j = 0; j < simcomp->columnnum; j++)
        simcomp->max_mcs->vlist[j] |= simcomp->max_mcs->sccs[i].vlist[j];
    }
  }
#ifdef DEBUG
if (simcomp->debug_mode % 2 == 1){
printf("recall extend_MCS\n");
}
#endif

  /* extend_MCSの前準備 */
  extract_Node(simcomp);

  /* 繰り返し max_MCS を伸ばしていく */
  while(extend_MCS(simcomp) != 0){
    ret = 1;
  }

  /* max_MCSのスコアを設定 */
  if (ret != 0) score_maxMCS(simcomp);

  return;
}

/* ノードとして構築されているが、クリークに使用されていないものを抽出 */
void extract_Node(SIMCOMP *simcomp){
  if (simcomp->max_mcs == NULL) return;

  switch(simcomp->node_mode){
  case _NT_ATOM_:
  case _NT_ATOM2_:
    extract_Node_byAtom(simcomp);
    break;
  case _NT_BOND_:
    extract_Node_byBond(simcomp);
    break;
  }

  return;
}

/*
 ノードに使用されている原子の used は YES
 そのうちmax_mcsに含まれていれば NO にする
*/
void extract_Node_byAtom(SIMCOMP *simcomp){
  int i;

  for (i = 0; i < simcomp->cpdmol_1->atomnum; i++)
    simcomp->cpdmol_1->atom[i].used = _NO_;
  for (i = 0; i < simcomp->cpdmol_2->atomnum; i++)
    simcomp->cpdmol_2->atom[i].used = _NO_;

  for (i = 0; i < simcomp->nodenum + simcomp->nodetrick_num; i++){
    simcomp->cpdmol_1->atom[simcomp->dnode[i].e1].used = _YES_;
    simcomp->cpdmol_2->atom[simcomp->dnode[i].e2].used = _YES_;
  }

  for (i = 0; i < simcomp->nodenum + simcomp->nodetrick_num; i++){
    if (getBit(simcomp->max_mcs->vlist, i) != 0){
      simcomp->cpdmol_1->atom[simcomp->dnode[i].e1].used = _NO_;
      simcomp->cpdmol_2->atom[simcomp->dnode[i].e2].used = _NO_;
    }
  }

  return;
}

/*
 ノードに使用されている結合の used は YES
 そのうちクリークに含まれていれば NO にする
*/
void extract_Node_byBond(SIMCOMP *simcomp){
  int i;

  for (i = 0; i < simcomp->cpdmol_1->bondnum; i++)
    simcomp->cpdmol_1->bond[i].used = _NO_;
  for (i = 0; i < simcomp->cpdmol_2->bondnum; i++)
    simcomp->cpdmol_2->bond[i].used = _NO_;

  for (i = 0; i < simcomp->nodenum + simcomp->nodetrick_num; i++){
    simcomp->cpdmol_1->bond[simcomp->dnode[i].e1].used = _YES_;
    simcomp->cpdmol_2->bond[simcomp->dnode[i].e2].used = _YES_;
  }

  for (i = 0; i < simcomp->nodenum + simcomp->nodetrick_num; i++){
    if (getBit(simcomp->max_mcs->vlist, i) != 0){
      simcomp->cpdmol_1->bond[simcomp->dnode[i].e1].used = _NO_;
      simcomp->cpdmol_2->bond[simcomp->dnode[i].e2].used = _NO_;
    }
  }

  return;
}

/*
  実際に広げることのできた個数を返す
*/
int extend_MCS(SIMCOMP *simcomp){
  int ret = 0;

  switch(simcomp->node_mode){
  case _NT_ATOM_:
  case _NT_ATOM2_:
    ret = extend_MCS_byAtom(simcomp);
    break;
  case _NT_BOND_:
    ret = extend_MCS_byBond(simcomp);
    break;
  }

  return ret;
}

int extend_MCS_byAtom(SIMCOMP *simcomp){
  int ret = 0;
  int i, j;

  /* 延長分を格納するので初期化 */
  for (i = 0; i < simcomp->columnnum; i++) simcomp->bitset1[i] = 0;

  /* ノードでループ */
  for (i = 0; i < simcomp->nodenum + simcomp->nodetrick_num; i++){

    /* ノードの原子の used が NO ならスキップ */
    if (simcomp->cpdmol_1->atom[simcomp->dnode[i].e1].used == _NO_) continue;
    if (simcomp->cpdmol_2->atom[simcomp->dnode[i].e2].used == _NO_) continue;

    for (j = 0; j < simcomp->columnnum; j++){
      /* ノードとクリーク中のノードが隣接しているか */
      if ((simcomp->connect[i][j] & simcomp->max_mcs->vlist[j]) != 0 ||
          (simcomp->connect[i][j] & simcomp->bitset1[j]) != 0){
        setBit(simcomp->bitset1, i);
        ret++;
        /* ノードの原子の used を NO に */
        simcomp->cpdmol_1->atom[simcomp->dnode[i].e1].used = _NO_;
        simcomp->cpdmol_2->atom[simcomp->dnode[i].e2].used = _NO_;
        break;
      }
    }
  }

  /* 延長分をマージ */
  for (i = 0; i < simcomp->columnnum; i++)
    simcomp->max_mcs->vlist[i] |= simcomp->bitset1[i];

  return ret;
}

int extend_MCS_byBond(SIMCOMP *simcomp){
  int ret = 0;
  int i, j;
  int ok;
  int idx1, idx2;

  /* 延長分を格納するので初期化 */
  for (i = 0; i < simcomp->columnnum; i++) simcomp->bitset1[i] = 0;

  /* ノードでループ */
  for (i = 0; i < simcomp->nodenum + simcomp->nodetrick_num; i++){
    /* ノードの結合の used が NO ならスキップ */
    if (simcomp->cpdmol_1->bond[simcomp->dnode[i].e1].used == _NO_) continue;
    if (simcomp->cpdmol_2->bond[simcomp->dnode[i].e2].used == _NO_) continue;

    ok = _NO_;
    for (j = 0; j < simcomp->columnnum; j++){
      /* ノードとクリーク中のノードが隣接しているか */
      if ((simcomp->connect[i][j] & simcomp->max_mcs->vlist[j]) != 0 ||
          (simcomp->connect[i][j] & simcomp->bitset1[j]) != 0){
        ok = _YES_;
        break;
      }
    }
    if (ok == _NO_) continue;

/* クリーク中の結合と共通原子の数が不一致なら、延長不可 */

    for (j = 0; j < simcomp->nodenum + simcomp->nodetrick_num; j++){
      /* クリークに含まれる結合について */
      if (getBit(simcomp->max_mcs->vlist, j) != 0 ||
          getBit(simcomp->bitset1, j) != 0){
        /* クリークのノードと未格納ノードの共通原子は(クエリー側) */
        idx1 = commonAtom(&(simcomp->cpdmol_1->bond[simcomp->dnode[i].e1]),
                          &(simcomp->cpdmol_1->bond[simcomp->dnode[j].e1]));

        /* クリークのノードと未格納ノードの共通原子は(ターゲット側) */
        idx2 = commonAtom(&(simcomp->cpdmol_2->bond[simcomp->dnode[i].e2]),
                          &(simcomp->cpdmol_2->bond[simcomp->dnode[j].e2]));

        if ((idx1 == -1 && idx2 != -1) || (idx1 != -1 && idx2 == -1)){
          ok = _NO_;
          break;
        }
      }
    }
    if (ok == _NO_) continue;

#ifdef DEBUG
if (simcomp->debug_mode % 2 == 1){
printf("add node in extend Node %d : Bond %d-%d\n", i + 1, simcomp->dnode[i].e1 + 1, simcomp->dnode[i].e2 + 1);
}
#endif
    setBit(simcomp->bitset1, i);
    ret++;
    /* ノードの結合の used を NO に */
    simcomp->cpdmol_1->bond[simcomp->dnode[i].e1].used = _NO_;
    simcomp->cpdmol_2->bond[simcomp->dnode[i].e2].used = _NO_;
  }

  /* 延長分をマージ */
  for (i = 0; i < simcomp->columnnum; i++)
    simcomp->max_mcs->vlist[i] |= simcomp->bitset1[i];

  return ret;
}

/* Alignment情報の構築　(alignnumのカウント) */
void make_alignInfo(SIMCOMP *simcomp){
  int i;

  if (simcomp->mcs_max == 0 || simcomp->max_mcs == NULL) return;

  /* 原子、結合のマッピング作成(キラルチェックしてあれば構築済み) */
  if (simcomp->chiral == _CC_NONE_)
    make_MAP(simcomp->max_mcs, simcomp);

  simcomp->alignnum = 0;
  for (i = 0; i < simcomp->cpdmol_1->atomnum; i++)
    if (simcomp->max_mcs->atommap[i] != 0) simcomp->alignnum++;

  return;
}

/*************************************
 ビット処理
*************************************/
/*
 指定位置にビットが立っている値を返す(範囲外が指定されていたら0x0を返す)
 毎回シフト演算を行うより、定数配列の値を返す方が良い気がする
*/
unsigned int bit(int pos){
  if (pos < 0 || pos > _BIT_LENGTH_) return bitflags[_BIT_LENGTH_];
  return bitflags[pos];
}

/* 指定位置のビットをONにする */
void setBit(unsigned int *array, int col){
  int colpos, bitpos;

  colpos = col / _BIT_LENGTH_;
  bitpos = col % _BIT_LENGTH_;
  array[colpos] |= bit(bitpos);

  return;
}

/* 指定位置のビットをOFFにする */
void resetBit(unsigned int *array, int col){
  int colpos, bitpos;

  colpos = col / _BIT_LENGTH_;
  bitpos = col % _BIT_LENGTH_;
  array[colpos] ^= bit(bitpos);

  return;
}

/* マトリックス中、指定位置のビットをONにする */
void setBitInMatrix(unsigned int **matrix, int row, int col){
  setBit(matrix[row], col);

  return;
}

/* 指定位置のビットを返す */
unsigned int getBit(unsigned int *array, int pos){
  int colpos, bitpos;

  colpos = pos / _BIT_LENGTH_;
  bitpos = pos % _BIT_LENGTH_;

  return array[colpos] & bit(bitpos);
}

/* ビット列における1の個数 */
/* 備　考   : java.util.BitSet.bitCount()　(since JDK 1.4)
             For a derivation of this algorithm, see
               "Algorithms and data structures with applications to 
                graphics and geometry", 
             by Jurg Nievergelt and Klaus Hinrichs, Prentice Hall, 1993.
*/
int bitCount(unsigned int val){
  val -= (val & 0xaaaaaaaa) >> 1;
  val =  (val & 0x33333333) + ((val >> 2) & 0x33333333);
  val =  (val + (val >> 4)) & 0x0f0f0f0f;
  val += val >> 8;

  return ((int)(val) + (int)(val >> 16)) & 0xff;
}

/*************************************
 メモリ
*************************************/
unsigned int **ctl_mallocArray_uint2(int row, int col)
{

  unsigned int **pp = NULL;
  int i = 0;

  /* 入力チェック */
  if (row < 1 || col < 1) {
    return pp;
  }

  /* int* の管理配列のメモリ領域を確保 */
  pp = (unsigned int**)malloc(sizeof(unsigned int*) * row);
  if (pp != NULL) {
    /* メモリ領域を初期化する */
    memset(pp, 0, sizeof(unsigned int*) * row);
  } else {
    /* NULLを返す */
    return pp;
  }

  for (i = 0; i < row; i++) {
    /* int の２次元のメモリ領域を確保 */
    unsigned int *p = (unsigned int*)malloc(sizeof(unsigned int) * col);

    /* unsigned int の２次元のメモリ領域を確保に成功したかどうか */
    if (p != NULL) {
      /* メモリ領域を初期化する */
      memset(p, 0, sizeof(unsigned int) * col);
      pp[i] = p;
    } else {
      int j = 0;

      /* 失敗する前までに確保したメモリ領域を開放 */
      for (j = 0; j < i; j++){
        free(pp[j]);
        pp[j] = NULL;
      }
      free(pp);
      pp = NULL;

      /* NULLを返す */
      return pp;
    }
  }

  return pp;
}

void ctl_freeArray_uint2(unsigned int ***pp, int row)
{
  int i = 0;

  if ((*pp) != NULL) {
    for (i = 0; i < row; i++) {
      if ((*pp)[i] != NULL) {
        free((*pp)[i]);
        (*pp)[i] = NULL;
      }
    }
    free(*pp);
    (*pp) = NULL;
  }

  return;
}

char* StrAllocCopy(char** dest, char* src)
{
  if (*dest != NULL){
    free(*dest);
  }
  if (src == NULL){
    *dest = NULL;
  } else {
    *dest = (char *) malloc (strlen(src) + 1);
    if (*dest == NULL) {
      *dest = NULL;
    } else {
      strcpy (*dest, src);
    }
  }
  return *dest;
}

/* 原子タイプ、原子クラス、元素記号の一致毎にペナルティを設定する */
/* 原子タイプ：0、原子クラス：-0.1、元素記号:-0.2 */
/* void CountAtomPenalty(SIMCOMP *simcomp) */
void CountAtomPenalty(SIMCOMP *simcomp) 
{
  int i;
  int idx1, idx2;
  char *ptr_t1, *ptr_t2, *ptr_c1, *ptr_c2, *ptr_a1, *ptr_a2;
  double atom_penalty = 0.0;
  int match_num_t, match_num_c, match_num_a;

  match_num_t = match_num_c = match_num_a = 0;

#ifdef DEBUG
if (simcomp->debug_mode % 2 == 1){
printf("call CountAtomPenalty\n");
}
#endif
  if(simcomp->max_mcs == NULL) return;

  simcomp->max_mcs->atom_penalty = 0.0;

  for(i=0; i < simcomp->cpdmol_1->atomnum; i++){
    if(simcomp->max_mcs->atommap == NULL) return;
    if(simcomp->max_mcs->atommap[i] != 0){
      idx1 = i;                           /* クエリの原子 */
      idx2 = simcomp->max_mcs->atommap[i]; /* ターゲットの原子(1から数える) */

      if(idx2 < 0) idx2 = -idx2;

      /* クエリ、ターゲットの原子タイプ */
      ptr_t1 = simcomp->cpdmol_1->atom[idx1].nid;
      ptr_t2 = simcomp->cpdmol_2->atom[idx2 - 1].nid;

      /* クエリ、ターゲットの原子class */
      ptr_c1 = simcomp->cpdmol_1->atom[idx1].cid;
      ptr_c2 = simcomp->cpdmol_2->atom[idx2 - 1].cid;
      
      /* クエリ、ターゲットの原子種 */
      ptr_a1 = simcomp->cpdmol_1->atom[idx1].aid;
      ptr_a2 = simcomp->cpdmol_2->atom[idx2 - 1].aid;

      if(strcmp(ptr_t1, ptr_t2) == 0){
        match_num_t++;
      }
      else if(strcmp(ptr_c1, ptr_c2) == 0){
        atom_penalty += 0.1;
        match_num_c++;
      }
      else if(strcmp(ptr_a1, ptr_a2) == 0){
        atom_penalty += 0.2;
        match_num_a++;
      }
      else{
        /* ここには入らないはず */
      }

#ifdef DEBUG
if (simcomp->debug_mode % 2 == 1){
printf("q: %4s, %4s, %4s, t: %4s, %4s, %4s\n",
ptr_t1, ptr_c1, ptr_a1,
ptr_t2, ptr_c2, ptr_a2);
}
#endif

    }
  }

/**/
#ifdef DEBUG
if (simcomp->debug_mode % 2 == 1){
/**/
printf("match_type, match_class, match_atom = %d, %d, %d\n",
match_num_t, match_num_c, match_num_a);
/**/
}
#endif
/**/

  simcomp->max_mcs->atom_penalty = atom_penalty;

}

