/*
  mcs.c
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "simcomp.h"

/* private function */
void make_SCCS(MCS *, SIMCOMP *);
void make_MAP_byAtom(MCS *, SIMCOMP *);
void make_MAP_byBond(MCS *, SIMCOMP *);
int check_chiral(MCS *, SIMCOMP *);
double calcRotation(double [][3]);
int check_geometry(MCS *, SIMCOMP *);
double torsionAngle(double *, double *, double *, double *);
void cross(double *, double *, double *);
double angle(double *, double *); 
double innerProduct(double *, double *);
void transBtoA(MCS *, SIMCOMP *);
MCS_BLOCK* create_MCS_BLOCK();
void freeMember_MCS(MCS *);
void free_MCS_BLOCK(MCS_BLOCK *);

/*
  MCS のスコアをつける
*/
void score_MCS_list(SIMCOMP *simcomp){
  int i, j;
  int lastnum;
  int ret;

  MCS_BLOCK *p;
  MCS *mcs;

#ifdef DEBUG
if (simcomp->debug_mode % 2 == 1){
printf("==mcs score== (%d)\n", simcomp->mcslist.num);
}
#endif
  p = simcomp->mcslist.pTop;
  while(p != NULL) {
    if (p->pNext != NULL){
      lastnum = 100;
    } else {
      lastnum = ((simcomp->mcslist.num - 1) % 100) + 1;
    }

    for (i = 0; i < lastnum; i++){
      mcs = &(p->mcs[i]);
      mcs->score = 0.0;
      for (j = 0; j < simcomp->nodenum + simcomp->nodetrick_num; j++){
        if (getBit(mcs->vlist, j) != 0)
          mcs->score += simcomp->dnode[j].score;
      }
      mcs->score +=
        (simcomp->cpdmol_1->nodenum + simcomp->cpdmol_2->nodenum - mcs->vnum * 2);

      /* ついでにSCCSもここで作成 */
      make_SCCS(mcs, simcomp);

      /* キラルチェック */
      ret = 0;
      if (simcomp->chiral != _CC_NONE_){
        /* 原子、結合のマッピング作成 */
#ifdef DEBUG
if (simcomp->debug_mode % 2 == 1){
printf("call make_MAPn");
}
#endif
        make_MAP(mcs, simcomp);
#ifdef DEBUG
if (simcomp->debug_mode % 2 == 1){
printf("ATOM MAP\n");
for (j = 0; j < simcomp->cpdmol_1->atomnum; j++){
printf("%2d - %2d\n", j + 1, mcs->atommap[j]);
}
}
#endif
        mcs->chiral_penalty = 0.0;
        if ((simcomp->chiral & _CC_CHIRAL_) != 0){
          /* 不斉チェック */
          /* この処理の途中から原子マッピングの値が負になるものが現れる */
#ifdef DEBUG
if (simcomp->debug_mode % 2 == 1){
printf("call check_chiral\n");
}
#endif
          ret += check_chiral(mcs, simcomp);
        }
        if ((simcomp->chiral & _CC_GEOMETRY_) != 0){
          /* trans/cisチェック */
          /* この処理の途中から結合マッピングの値が負になるものが現れる */
#ifdef DEBUG
if (simcomp->debug_mode % 2 == 1){
printf("call check_geometry\n");
}
#endif
          ret += check_geometry(mcs, simcomp);
        }

        /* スコアの補正 */
        mcs->score += (double)ret;

        /* 結合の不一致情報を原子に移行(出力情報のため) */
#ifdef DEBUG
if (simcomp->debug_mode % 2 == 1){
printf("call transBtoA\n");
}
#endif
        transBtoA(mcs, simcomp);
      }

#ifdef DEBUG
if (simcomp->debug_mode % 2 == 1){
printf("vlist");
for (j = 0; j < simcomp->columnnum; j++){
printf(" %08X", mcs->vlist[j]);
}
printf(" : score = %5.2f(%d) : sccs", mcs->score, ret);
for (j = 0; j < mcs->sccsnum; j++){
printf(" (%d, %5.2f)", mcs->sccs[j].vnum, mcs->sccs[j].score);
}
printf("\n");
}
#endif
    }
    p = p->pNext;
  }

  return;
}

/*
  max_MCS のスコアをつける
*/
void score_maxMCS(SIMCOMP *simcomp){
  int i;

  if (simcomp->max_mcs == NULL) return;

  simcomp->max_mcs->vnum = 0;
  simcomp->max_mcs->score = 0.0;
  for (i = 0; i < simcomp->nodenum + simcomp->nodetrick_num; i++){
    if (getBit(simcomp->max_mcs->vlist, i) != 0){
      simcomp->max_mcs->vnum++;
      simcomp->max_mcs->score += simcomp->dnode[i].score;
    }
  }
  simcomp->max_mcs->score +=
    (simcomp->cpdmol_1->nodenum + simcomp->cpdmol_2->nodenum - simcomp->max_mcs->vnum * 2);

  /* ついでにSCCSもここで作成 */
  make_SCCS(simcomp->max_mcs, simcomp);
  simcomp->mcs_max = simcomp->max_mcs->vnum;

  /* キラルチェック */
  if (simcomp->chiral != _CC_NONE_){
    /* 原子、結合のマッピング作成 */
    /* extendにより状況が変化するので再定義 */
    make_MAP(simcomp->max_mcs, simcomp);

    simcomp->max_mcs->chiral_penalty = 0.0;
    if ((simcomp->chiral & _CC_CHIRAL_) != 0){
      /* 不斉チェック */
      /* この処理の途中から原子マッピングの値が負になるものが現れる */
      check_chiral(simcomp->max_mcs, simcomp);
    }
    if ((simcomp->chiral & _CC_GEOMETRY_) != 0){
      /* trans/cisチェック */
      /* この処理の途中から結合マッピングの値が負になるものが現れる */
      check_geometry(simcomp->max_mcs, simcomp);
    }

    /* 結合の不一致情報を原子に移行(出力情報のため) */
    transBtoA(simcomp->max_mcs, simcomp);
  }

  return;
}
/*
  MCS に SCCS を作る
  ここでSCCSの領域は最初のvnum分
  extend_MCS()実行後に再設定のため呼ばれるが、SCCSの数は増加しないはず
*/
void make_SCCS(MCS *mcs, SIMCOMP *simcomp){
  int i, j;
  int diff_sum = 0; /* ビット列配列の0でない個数 */
  int mask_sum = 0; /* ビット列配列の0でない個数 */
  unsigned int *maskbit = simcomp->bitset1;  /* SCCSにアサインしないといけない場所 */
  unsigned int *vlist   = simcomp->bitset2;  /* SCCS */
  unsigned int *before  = simcomp->bitset3;  /* マージ前 */
  unsigned int *diff    = simcomp->bitset4;  /* マージ前後の差分 */
  int max;
  int ti;
  unsigned int tui;
  double td;

  /* extend_MCS()実行後に再設定の場合、一度クリア(vnumが増えている) */
  if (mcs->sccs != NULL){
    for (i = 0; i < mcs->sccsnum; i++){
      if (mcs->sccs[i].vlist != NULL){
        free(mcs->sccs[i].vlist);
      }
    }
    free(mcs->sccs);
    mcs->sccs = NULL;
    mcs->sccsnum = 0;
  }

  /* ここでSCCSの領域を確保(最大でもvnum分) */
  mcs->sccs = (SCCS *)malloc(sizeof(SCCS) * mcs->vnum);
  if (mcs->sccs == NULL){
    return;
  } else {
    memset(mcs->sccs, 0, sizeof(SCCS) * mcs->vnum);
  }

  for (i = 0; i < simcomp->columnnum; i++){
    maskbit[i] = mcs->vlist[i];
    if (maskbit[i] != 0) mask_sum++;
  }
  while(mask_sum > 0){

    /* クリア */
    for (i = 0; i < simcomp->columnnum; i++){
      vlist[i] = 0;
      diff[i] = 0;
    }

    /* maskbitの最初の1ビットをSCCSに */
    for (i = 0; i < simcomp->nodenum + simcomp->nodetrick_num; i++){
      if (getBit(maskbit, i) != 0){
        setBit(vlist, i);
        setBit(diff, i);
        diff_sum = 1;
        break;
      }
    }

    /* SCCSに追加したデータから芋づる式に追加 */
    while(diff_sum > 0){
      /* 更新前のビット列 */
      for (i = 0; i < simcomp->columnnum; i++) before[i] = vlist[i];

      for (i = 0; i < simcomp->nodenum + simcomp->nodetrick_num; i++){
        if (getBit(diff, i) != 0){
          for (j = 0; j < simcomp->columnnum; j++){
            vlist[j] |= simcomp->connect[i][j];
            vlist[j] &= maskbit[j];
          }
        }
      }

      diff_sum = 0;
      for (i = 0; i < simcomp->columnnum; i++){
        diff[i] = vlist[i] ^ before[i];
        if (diff[i] != 0) diff_sum++;
      }

    }

    /* このときのvlistがSCCSなので追加 */
    mcs->sccs[mcs->sccsnum].vnum = 0;
    mcs->sccs[mcs->sccsnum].score = 0.0;
    mcs->sccs[mcs->sccsnum].vlist = 
          (unsigned int*)malloc(sizeof(unsigned int) * simcomp->columnnum);
    memset(mcs->sccs[mcs->sccsnum].vlist, 0, sizeof(unsigned int) * simcomp->columnnum);
    for (i = 0; i < simcomp->nodenum + simcomp->nodetrick_num; i++){
      if (getBit(vlist, i) != 0){
        mcs->sccs[mcs->sccsnum].vnum++;
        mcs->sccs[mcs->sccsnum].score += simcomp->dnode[i].score;
      }
    }
    for (i = 0; i < simcomp->columnnum; i++){
      mcs->sccs[mcs->sccsnum].vlist[i] = vlist[i];
    }
    mcs->sccsnum++;

    /* SCCSにアサインしていない残りの場所を更新 */
    mask_sum = 0;
    for (i = 0; i < simcomp->columnnum; i++){
      maskbit[i] ^= vlist[i];
      if (maskbit[i] != 0) mask_sum++;
    }
  }

  /* SCCSリストを降順にソート */
  /* データ数が十分に少ないだろうからselection sort */
  for (i = 0; i < mcs->sccsnum - 1; i++){
    max = i;
    for (j = i + 1; j < mcs->sccsnum; j++){
      if (mcs->sccs[max].vnum < mcs->sccs[j].vnum ||
         (mcs->sccs[max].vnum == mcs->sccs[j].vnum &&
          mcs->sccs[max].score < mcs->sccs[j].score))
        max = j;
    }

    if (i != max){
      ti = mcs->sccs[i].vnum;
      mcs->sccs[i].vnum = mcs->sccs[max].vnum;
      mcs->sccs[max].vnum = ti;
      td = mcs->sccs[i].score;
      mcs->sccs[i].score = mcs->sccs[max].score;
      mcs->sccs[max].score = td;
      for (j = 0; j < simcomp->columnnum; j++){
        tui = mcs->sccs[i].vlist[j];
        mcs->sccs[i].vlist[j] = mcs->sccs[max].vlist[j];
        mcs->sccs[max].vlist[j] = tui;
      }
    }
  }

  return;
}

/*
  MCS 原子、結合のマッピングを作成(キラルチェックに使用)
*/
void make_MAP(MCS *mcs, SIMCOMP *simcomp){
  if (mcs == NULL) return;

  if (mcs->atommap == NULL){  /* extendの後にもここに来る */
    mcs->atommap = (int*)malloc(sizeof(int) * simcomp->cpdmol_1->atomnum);
    if (mcs->atommap == NULL) {
      return;
    }
  }
  memset(mcs->atommap, 0, sizeof(int) * simcomp->cpdmol_1->atomnum);

  if (mcs->bondmap == NULL){  /* extendの後にもここに来る */
    mcs->bondmap = (int*)malloc(sizeof(int) * simcomp->cpdmol_1->bondnum);
    if (mcs->bondmap == NULL) {
      return;
    }
  }
  memset(mcs->bondmap, 0, sizeof(int) * simcomp->cpdmol_1->bondnum);

  switch(simcomp->node_mode){
  case _NT_ATOM_:
  case _NT_ATOM2_:
    make_MAP_byAtom(mcs, simcomp);
    break;
  case _NT_BOND_:
    make_MAP_byBond(mcs, simcomp);
    break;
  }
  return;
}

/*
  MCS 原子、結合のマッピングを作成(Nodeが原子の場合)
*/
void make_MAP_byAtom(MCS *mcs, SIMCOMP *simcomp){
  int i, j;
  int a11, a12;
  int a21, a22;

  for (i = 0; i < simcomp->cpdmol_1->atomnum; i++)
    mcs->atommap[i] = 0;

  /* ドッキングノードから原子情報 */
  for (i = 0; i < mcs->sccsnum; i++){
    for (j = 0; j < simcomp->nodenum + simcomp->nodetrick_num ; j++){
      if (getBit(mcs->sccs[i].vlist, j) != 0){
        /* 相手の番号を格納(1から数える) */
        mcs->atommap[simcomp->dnode[j].e1] = simcomp->dnode[j].e2 + 1;
      }
    }
  }

  /* 結合情報は両端原子のSCCS成分が一致しないといけない */
  for (i = 0; i < simcomp->cpdmol_1->bondnum; i++){
    /* クエリー側 */
    a11 = simcomp->cpdmol_1->bond[i].a1 - 1;
    a12 = simcomp->cpdmol_1->bond[i].a2 - 1;
    /* ターゲット側(1から数える) */
    a21 = mcs->atommap[a11];
    a22 = mcs->atommap[a12];

    /* 相手が居ないならばスキップ */
    if (a21 == 0 || a22 == 0) continue;

    /* 相手の結合番号(1から数える) */
    for (j = 0; j < simcomp->cpdmol_2->bondnum; j++){
      if ((a21 == simcomp->cpdmol_2->bond[j].a1 &&
           a22 == simcomp->cpdmol_2->bond[j].a2) ||
          (a21 == simcomp->cpdmol_2->bond[j].a2 &&
           a22 == simcomp->cpdmol_2->bond[j].a1)){
         mcs->bondmap[i] = j + 1;
         break;
      }
    }
  }
  return;
}

/*
  MCS 原子、結合のマッピングを作成(Nodeが結合の場合)
*/
void make_MAP_byBond(MCS *mcs, SIMCOMP *simcomp){
  int i, j, k;
  int ok;
  int b1, b2;
  BOND *bond1, *bond2;
  ATOM *q1, *q2, *t1, *t2;
  char *qid, *tid;
  int score1, score2;

  /* 原子とドッキングノードの対応関係初期化 */
  for (i = 0; i < simcomp->cpdmol_1->atomnum; i++)
    for (j = 0; j < simcomp->columnnum; j++)
      simcomp->table1[i][j] = 0;
  for (i = 0; i < simcomp->cpdmol_2->atomnum; i++)
    for (j = 0; j < simcomp->columnnum; j++)
      simcomp->table2[i][j] = 0;
  for (i = 0; i < simcomp->cpdmol_1->atomnum; i++)
    mcs->atommap[i] = 0;
  for (i = 0; i < simcomp->cpdmol_1->bondnum; i++)
    mcs->bondmap[i] = 0;

  /* ドッキングノードから結合情報 */
  for (i = 0; i < mcs->sccsnum; i++){
    for (j = 0; j < simcomp->nodenum + simcomp->nodetrick_num ; j++){
      if (getBit(mcs->sccs[i].vlist, j) != 0){
        b1 = simcomp->dnode[j].e1;
        b2 = simcomp->dnode[j].e2;
        bond1 = &(simcomp->cpdmol_1->bond[b1]);
        bond2 = &(simcomp->cpdmol_2->bond[b2]);
        /* 相手の結合番号(1から数える) */
        mcs->bondmap[b1] = b2 + 1;

        if (mcs->sccs[i].vnum == 1){
          /* 結合1個分のSCCS　どっち向きがいいかをチェックして設定 */

          q1 = &(simcomp->cpdmol_1->atom[bond1->a1 - 1]);
          q2 = &(simcomp->cpdmol_1->atom[bond1->a2 - 1]);
          t1 = &(simcomp->cpdmol_2->atom[bond2->a1 - 1]);
          t2 = &(simcomp->cpdmol_2->atom[bond2->a2 - 1]);

          switch(simcomp->atom_mode){
          case _DC_NORMAL_:
            qid = q1->nid;
            tid = t1->nid;
            break;
          case _DC_CLASS_:
            qid = q1->cid;
            tid = t1->cid;
            break;
          case _DC_ATOM_:
            qid = q1->aid;
            tid = t1->aid;
            break;
          default:
            qid = q1->nid;
            tid = t1->nid;
            break;
          }
          /* 結合をa1とa1、a2とa2で重ね合わせ状況 */
          score1 = 0;
          score2 = 0;
          if (q1->RAneighbor == _YES_ || t1->RAneighbor == _YES_){
            if (strcmp(q1->aid, t1->aid) == 0) score1 += 1;
            else                               score1 += -999;
          } else {
            if (strcmp(qid, tid) == 0)         score1 += 2;
            else                               score1 += -999;
          }
          if (q2->RAneighbor == _YES_ || t2->RAneighbor == _YES_){
            if (strcmp(q2->aid, t2->aid) == 0) score1 += 1;
            else                               score1 += -999;
          } else {
            if (strcmp(qid, tid) == 0)         score1 += 2;
            else                               score1 += -999;
          }

          /* 結合をa1とa2、a2とa1で重ね合わせ状況 */
          if (q1->RAneighbor == _YES_ || t2->RAneighbor == _YES_){
            if (strcmp(q1->aid, t2->aid) == 0) score2 += 1;
            else                               score2 += -999;
          } else {
            if (strcmp(qid, tid) == 0)         score2 += 2;
            else                               score2 += -999;
          }
          if (q2->RAneighbor == _YES_ || t1->RAneighbor == _YES_){
            if (strcmp(q2->aid, t1->aid) == 0) score2 += 1;
            else                               score2 += -999;
          } else {
            if (strcmp(qid, tid) == 0)         score2 += 2;
            else                               score2 += -999;
          }

          if (score1 > 0 || score2 > 0){  /* ともに負の値なら重ね合わせ不可 */
            if (score1 >= score2){    /* a1とa1、a2とa2 */
              /* 相手の番号を格納(1から数える) */
              mcs->atommap[bond1->a1 - 1] = bond2->a1;
              mcs->atommap[bond1->a2 - 1] = bond2->a2;
            } else {                  /* a1とa2、a2とa1 */
              /* 相手の番号を格納(1から数える) */
              mcs->atommap[bond1->a1 - 1] = bond2->a2;
              mcs->atommap[bond1->a2 - 1] = bond2->a1;
            }
          }
          break;
        } else {
          /* ドッキングノード番号をセット */
          setBit(simcomp->table1[bond1->a1 - 1], j);
          setBit(simcomp->table1[bond1->a2 - 1], j);
          setBit(simcomp->table2[bond2->a1 - 1], j);
          setBit(simcomp->table2[bond2->a2 - 1], j);
        }
      }
    }
  }

  /*
     クエリーとターゲットの原子を比較
     原子と結合ノードとの対応関係が一致していれば重ね合わせ可能
   */
  for (i = 0; i < simcomp->cpdmol_1->atomnum; i++){
    if (mcs->atommap[i] != 0) continue;

    ok = _NO_;
    for (j = 0; j < simcomp->columnnum; j++){
      if (simcomp->table1[i][j] != 0){
        ok = _YES_;
        break;
      }
    }
    if (ok == _NO_) continue;

    for (j = 0; j < simcomp->cpdmol_2->atomnum; j++){
      ok = _YES_;
      for (k = 0; k < simcomp->columnnum; k++){
        if (simcomp->table1[i][k] != simcomp->table2[j][k]){
          ok = _NO_;
          break;
        }
      }
      if (ok == _YES_){  /* クリークとの関連が完全一致なら */
        mcs->atommap[i] = j + 1;
        break;
      }
    }
  }
  return;
}

/* 不斉チェック */
/* この処理の途中から原子マッピングの値が負になるものが現れる */
int check_chiral(MCS *mcs, SIMCOMP *simcomp){
  int i, j, k;
  int b;
  int qa[30][2], ta[30][2];  /* 不斉中心と結合相手　クエリーとターゲットの対応関係 */
  int wa[30][2];
  int ok;
  int count;
  int undef[2];
  int ret = 0;
  BOND *bond;
  double vec[4][3];
  double valQ, valT;

  for (i = 0; i < simcomp->cpdmol_1->atomnum; i++){
    qa[0][0] = i;
    ta[0][0] = abs(mcs->atommap[i]) - 1;
    qa[0][1] = 0;
    ta[0][1] = 0;

    /* 互いに不斉中心の候補であること */
    if (simcomp->cpdmol_1->atom[i].chiral == _NO_) continue;
    if (ta[0][0] == -1) continue;
    if (simcomp->cpdmol_2->atom[ta[0][0]].chiral == _NO_) continue;
#ifdef DEBUG
if (simcomp->debug_mode % 2 == 1){
printf("atom1(%2d) chiral %d -> atom2(%2d) chiral %d\n", i + 1, simcomp->cpdmol_1->atom[i].chiral, ta[0][0] + 1, simcomp->cpdmol_2->atom[ta[0][0]].chiral);
}
#endif

    for (j = 1; j < 30; j++){
      qa[j][0] = -1;
      qa[j][1] =  0;
      ta[j][0] = -1;
      ta[j][1] =  0;
    }
    for (j = 0; j < 30; j++){
      wa[j][0] = -1;
      wa[j][1] =  0;
    }
    /* クエリー側の結合相手 */
    for (j = 0; j < simcomp->cpdmol_1->atom[i].bondnum; j++){
      b = simcomp->cpdmol_1->atom[i].bond[j];
      bond = &(simcomp->cpdmol_1->bond[b]);
      if (bond->a1 == i + 1){
        qa[j + 1][0] = bond->a2 - 1;
        if (bond->prop != NULL){
          if (strcasecmp(bond->prop, "#UP") == 0){
            qa[j + 1][1] =  1;
          } else
          if (strcasecmp(bond->prop, "#DOWN") == 0){
            qa[j + 1][1] = -1;
          }
        }
      }
      if (bond->a2 == i + 1){
        qa[j + 1][0] = bond->a1 - 1;
        if (bond->prop != NULL){
          if (strcasecmp(bond->prop, "#UP") == 0){
            qa[j + 1][1] = -1;
          } else
          if (strcasecmp(bond->prop, "#DOWN") == 0){
            qa[j + 1][1] =  1;
          }
        }
      }
    }
    /* ターゲット側の結合相手をワークに */
    for (j = 0; j < simcomp->cpdmol_2->atom[ta[0][0]].bondnum; j++){
      b = simcomp->cpdmol_2->atom[ta[0][0]].bond[j];
      bond = &(simcomp->cpdmol_2->bond[b]);
      if (bond->a1 == ta[0][0] + 1){
        wa[j][0] = bond->a2 - 1;
        if (bond->prop != NULL){
          if (strcasecmp(bond->prop, "#UP") == 0){
            wa[j][1] =  1;
          } else
          if (strcasecmp(bond->prop, "#DOWN") == 0){
            wa[j][1] = -1;
          }
        }
      }
      if (bond->a2 == ta[0][0] + 1){
        wa[j][0] = bond->a1 - 1;
        if (bond->prop != NULL){
          if (strcasecmp(bond->prop, "#UP") == 0){
            wa[j][1] = -1;
          } else
          if (strcasecmp(bond->prop, "#DOWN") == 0){
            wa[j][1] =  1;
          }
        }
      }
    }
#ifdef DEBUG
if (simcomp->debug_mode % 2 == 1){
for (j = 0; j < 4; j++){
printf("w(%2d, %2d)\n", wa[j][0] + 1, wa[j][1]);
}
}
#endif

    /* ターゲット側の対応原子をセット */
    count = 0;
    undef[0] = -1;
    undef[1] =  0;
    for (j = 1; j < 5; j++){
      if (qa[j][0] == -1) continue;
      ta[j][0] = abs(mcs->atommap[qa[j][0]]) - 1;
      if (ta[j][0] == -1) continue;
      ok = _NO_;
      for (k = 0; k < 4; k++){
        if (ta[j][0] == wa[k][0]){  /* 対応原子が確定 */
          ok = _YES_;
          wa[k][0] = -1;
          ta[j][1] = wa[k][1];
          count++;
          break;
        }
      }
      if (ok == _NO_){
        ta[j][0] = -1;       /* 対応関係が見つからないとする */
      }
    }
    for (j = 0; j < 4; j++){
      if (wa[j][0] != -1){
        undef[0] = wa[j][0]; /* 余りものを記憶 */
        undef[1] = wa[j][1];
        break;
      }
    }
#ifdef DEBUG
if (simcomp->debug_mode % 2 == 1){
printf("u(%2d, %2d)\n", undef[0] + 1, undef[1]);
printf("count = %d\n", count);
}
#endif

    /* 互いの結合相手の内3つ以上で対応関係がなければ　立体は評価出来ない */
    if (count < 3) continue;

    /*
       結合相手の対応関係
             4つ判明 → そのまま立体の評価
             3つ判明 → 余ったもの同士が対応するとみなす
                        ターゲット側の結合数が3なら残りは省略された水素
     */
    if (count == 3 && undef[0] != -1){
      for (j = 1; j < 5; j++){
        if (ta[j][0] == -1){
          ta[j][0] = undef[0];
          ta[j][1] = undef[1];
          break;
        }
      }
    }
#ifdef DEBUG
if (simcomp->debug_mode % 2 == 1){
for (j = 0; j < 5; j++){
printf("q(%2d, %2d) t(%2d, %2d)\n", qa[j][0] + 1, qa[j][1], ta[j][0] + 1, ta[j][1]);
}
}
#endif

    /* クエリー側二面体角 */
    /* 4つのベクトルを設定 */
    count = 0;
    undef[0] = -1;
    for (j = 0; j < 4; j++){
      vec[j][0] = 0.0;
      vec[j][1] = 0.0;
      vec[j][2] = 0.0;
    }
    for (j = 0; j < 4; j++){
      if (qa[j + 1][0] == -1){ /* 省略された水素なので後で決定 */
        undef[0] = j;
        count++;
      } else {
        vec[j][0] = simcomp->cpdmol_1->atom[qa[j + 1][0]].x -
                  simcomp->cpdmol_1->atom[qa[0][0]].x;
        vec[j][1] = simcomp->cpdmol_1->atom[qa[j + 1][0]].y -
                  simcomp->cpdmol_1->atom[qa[0][0]].y;
        vec[j][2] = (double)qa[j + 1][1];
      }
    }
    if (count > 1) continue;
    if (count == 1){ /* 省略された水素 */
      for (j = 0; j < 4; j++){
        if (j == undef[0]) continue;
        vec[undef[0]][0] += vec[j][0];
        vec[undef[0]][1] += vec[j][1];
        vec[undef[0]][2] += vec[j][2];
      }
      vec[undef[0]][0] /= -3.0;
      vec[undef[0]][1] /= -3.0;
      vec[undef[0]][2] /= -3.0;
    }
    /* 二面体角 */
    valQ = calcRotation(vec);
    if (fabs(valQ) < NEARLY_ZERO) valQ = 0.0;

    /* ターゲット側二面体角 */
    /* 4つのベクトルを設定 */
    count = 0;
    undef[0] = -1;
    for (j = 0; j < 4; j++){
      vec[j][0] = 0.0;
      vec[j][1] = 0.0;
      vec[j][2] = 0.0;
    }
    for (j = 0; j < 4; j++){
      if (ta[j + 1][0] == -1){ /* 省略された水素なので後で決定 */
        undef[0] = j;
        count++;
      } else {
        vec[j][0] = simcomp->cpdmol_2->atom[ta[j + 1][0]].x -
                  simcomp->cpdmol_2->atom[ta[0][0]].x;
        vec[j][1] = simcomp->cpdmol_2->atom[ta[j + 1][0]].y -
                  simcomp->cpdmol_2->atom[ta[0][0]].y;
        vec[j][2] = (double)ta[j + 1][1];
      }
    }
    if (count > 1) continue;
    if (count == 1){ /* 省略された水素 */
      for (j = 0; j < 4; j++){
        if (j == undef[0]) continue;
        vec[undef[0]][0] += vec[j][0];
        vec[undef[0]][1] += vec[j][1];
        vec[undef[0]][2] += vec[j][2];
      }
      vec[undef[0]][0] /= -3.0;
      vec[undef[0]][1] /= -3.0;
      vec[undef[0]][2] /= -3.0;
    }
    /* 二面体角 */
    valT = calcRotation(vec);
    if (fabs(valT) < NEARLY_ZERO) valT = 0.0;

#ifdef DEBUG
if (simcomp->debug_mode % 2 == 1){
printf("q %10.4lf : t %10.4lf\n", valQ, valT);
}
#endif
    /* 二面体角の符号が一致するならスコアに加点 */
    if (valQ * valT > NEARLY_ZERO){
      ret++;
    } else
    /* 二面体角の符号が不一致なら原子マッピングにマーク */
    if (valQ * valT < - NEARLY_ZERO){
      mcs->atommap[i] *= -1;
      mcs->chiral_penalty += 0.1;
    }
  }
#ifdef DEBUG
if (simcomp->debug_mode % 2 == 1){
printf("additional chiral = %d\n", ret);
}
#endif
  return ret;
}

/* 不斉中心原子から各原子へ、4つのベクトルから回転の向き(±)を計算する */
/* 数値演算ライブラリを使用しないので、二面体角を計算するより楽な方法 */
double calcRotation(double v[][3]){
  double mat[3][3];

  mat[0][0] = v[1][0] - v[0][0];
  mat[0][1] = v[1][1] - v[0][1];
  mat[0][2] = v[1][2] - v[0][2];
  mat[1][0] = v[2][0] - v[0][0];
  mat[1][1] = v[2][1] - v[0][1];
  mat[1][2] = v[2][2] - v[0][2];
  mat[2][0] = v[3][0] - v[0][0];
  mat[2][1] = v[3][1] - v[0][1];
  mat[2][2] = v[3][2] - v[0][2];

  return mat[0][0] * (mat[1][1] * mat[2][2] - mat[1][2] * mat[2][1]) +
         mat[0][1] * (mat[1][2] * mat[2][0] - mat[1][0] * mat[2][2]) +
         mat[0][2] * (mat[1][0] * mat[2][1] - mat[1][1] * mat[2][0]);
}

/* trans/cisチェック */
/* この処理の途中から結合マッピングの値が負になるものが現れる */
int check_geometry(MCS *mcs, SIMCOMP *simcomp){
  int i, j;
  int qa[4], ta[4];  /* [0]-[1]=[2]-[3]　となるように結合している */
  int wa[2], wb[2];
  int tb;
  int ret = 0;
  int ok;
  BOND *bond = NULL;
  ATOM *atom = NULL;
  double c1[3], c2[3], c3[3], c4[3];
  double valQ, valT;

  for (i = 0; i < simcomp->cpdmol_1->bondnum; i++){
    bond = &(simcomp->cpdmol_1->bond[i]);
    tb = abs(mcs->bondmap[i]) - 1;
		/* add 20090727 */
		if(tb == -1) continue;
		/* add end */
    qa[1] = bond->a1 - 1;
    qa[2] = bond->a2 - 1;
    ta[1] = abs(mcs->atommap[qa[1]]) - 1;
    ta[2] = abs(mcs->atommap[qa[2]]) - 1;
    qa[0] = qa[3] = ta[0] = ta[3] = -1;

    /* 互いにtrans/cisの候補であること */
    if (bond->geometry == _NO_) continue;
    if (simcomp->cpdmol_2->bond[tb].geometry == _NO_) continue;
    /* 結合で対応関係にあるなら、原子のマッピングが不一致になならない */

#ifdef DEBUG
if (simcomp->debug_mode % 2 == 1){
printf("bond1(%2d) geometry %d -> bond2(%2d) geometry %d\n", i + 1, bond->geometry, tb + 1, simcomp->cpdmol_2->bond[tb].geometry);
}
#endif
    /* [0]-[1]の結合候補が、1つか2つあるはず */
    wb[0] = wb[1] = -1;
    atom = &(simcomp->cpdmol_1->atom[qa[1]]);
    for (j = 0; j < atom->bondnum; j++){
      if (atom->bond[j] == i) continue;
      if (wb[0] < 0) wb[0] = atom->bond[j];
      else wb[1] = atom->bond[j];
    }

    ok = _NO_;
    /* 結合候補その1がマッピングされているならセット */
    if (wb[0] != -1){
      tb = abs(mcs->bondmap[wb[0]]) - 1;
      if (tb != -1){
        wa[0] = simcomp->cpdmol_1->bond[wb[0]].a1 - 1;
        if (wa[0] == qa[1]) wa[0] = simcomp->cpdmol_1->bond[wb[0]].a2 - 1;
        qa[0] = wa[0];
        ta[0] = abs(mcs->atommap[qa[0]]) - 1;
        ok = _YES_;
      }
    }
    /* 結合候補その2がマッピングされているならセット */
    if (ok == _NO_ && wb[1] != -1){
      tb = abs(mcs->bondmap[wb[1]]) - 1;
      if (tb != -1){
        wa[1] = simcomp->cpdmol_1->bond[wb[1]].a1 - 1;
        if (wa[1] == qa[1]) wa[1] = simcomp->cpdmol_1->bond[wb[1]].a2 - 1;
        qa[0] = wa[1];
        ta[0] = abs(mcs->atommap[qa[0]]) - 1;
        ok = _YES_;
      }
    }
    /* 結合候補が見つからないならチェックしないでスルー */
    if (ok == _NO_) continue;

    /* [2]-[3]の結合候補が、1つか2つあるはず */
    wb[0] = wb[1] = -1;
    atom = &(simcomp->cpdmol_1->atom[qa[2]]);
    for (j = 0; j < atom->bondnum; j++){
      if (atom->bond[j] == i) continue;
      if (wb[0] < 0) wb[0] = atom->bond[j];
      else wb[1] = atom->bond[j];
    }

    ok = _NO_;
    /* 結合候補その1がマッピングされているならセット */
    if (wb[0] != -1){
      tb = abs(mcs->bondmap[wb[0]]) - 1;
      if (tb != -1){
        wa[0] = simcomp->cpdmol_1->bond[wb[0]].a1 - 1;
        if (wa[0] == qa[2]) wa[0] = simcomp->cpdmol_1->bond[wb[0]].a2 - 1;
        qa[3] = wa[0];
        ta[3] = abs(mcs->atommap[qa[3]]) - 1;
        ok = _YES_;
      }
    }
    /* 結合候補その2がマッピングされているならセット */
    if (ok == _NO_ && wb[1] != -1){
      tb = abs(mcs->bondmap[wb[1]]) - 1;
      if (tb != -1){
        wa[1] = simcomp->cpdmol_1->bond[wb[1]].a1 - 1;
        if (wa[1] == qa[2]) wa[1] = simcomp->cpdmol_1->bond[wb[1]].a2 - 1;
        qa[3] = wa[1];
        ta[3] = abs(mcs->atommap[qa[3]]) - 1;
        ok = _YES_;
      }
    }

#ifdef DEBUG
if (simcomp->debug_mode % 2 == 1){
printf("%2d - %2d = %2d - %2d: %2d - %2d = %2d - %2d\n", qa[0] + 1, qa[1] + 1, qa[2] + 1, qa[3] + 1, ta[0] + 1, ta[1] + 1, ta[2] + 1, ta[3] + 1);
}
#endif

    /* 結合候補が見つからないなら　スルー */
    if (ok == _NO_) continue;

    /*
       4原子の二面体角を求める
       二次元描画を前提としZ座標は0とする
       角度は0degか180degのどちらかのはず
       90degより小さければ、zusammen
       90degより大きければ、entgegen
     */
    c1[0] = simcomp->cpdmol_1->atom[qa[0]].x;
    c1[1] = simcomp->cpdmol_1->atom[qa[0]].y;
    c1[2] = 0.0;
    c2[0] = simcomp->cpdmol_1->atom[qa[1]].x;
    c2[1] = simcomp->cpdmol_1->atom[qa[1]].y;
    c2[2] = 0.0;
    c3[0] = simcomp->cpdmol_1->atom[qa[2]].x;
    c3[1] = simcomp->cpdmol_1->atom[qa[2]].y;
    c3[2] = 0.0;
    c4[0] = simcomp->cpdmol_1->atom[qa[3]].x;
    c4[1] = simcomp->cpdmol_1->atom[qa[3]].y;
    c4[2] = 0.0;
    valQ = torsionAngle(c1, c2, c3, c4);

    c1[0] = simcomp->cpdmol_2->atom[ta[0]].x;
    c1[1] = simcomp->cpdmol_2->atom[ta[0]].y;
    c2[0] = simcomp->cpdmol_2->atom[ta[1]].x;
    c2[1] = simcomp->cpdmol_2->atom[ta[1]].y;
    c3[0] = simcomp->cpdmol_2->atom[ta[2]].x;
    c3[1] = simcomp->cpdmol_2->atom[ta[2]].y;
    c4[0] = simcomp->cpdmol_2->atom[ta[3]].x;
    c4[1] = simcomp->cpdmol_2->atom[ta[3]].y;
    valT = torsionAngle(c1, c2, c3, c4);

    /* 計算エラー(3原子が直線上に並んだ場合)　スルー */
    if (valQ > 900.0) continue;
    if (valT > 900.0) continue;

    if ((valQ - PI_HALF) * (valT - PI_HALF) > 0.0){
      /* 90degでの大小関係　一致 */
      ret++;
    } else {
      /* 90degでの大小関係　不一致 */
      mcs->bondmap[i] *= -1;
      mcs->chiral_penalty += 0.1;
    }
  }
#ifdef DEBUG
if (simcomp->debug_mode % 2 == 1){
printf("additional geometry = %d\n", ret);

}
#endif
  return ret;
}

/* 二面体角を求める */
double torsionAngle(double *a, double *b, double *c, double *d){
  int i;
  double torsion;
  double b1[3], b2[3], b3[3];
  double c1[3], c2[3], c3[3];

  for (i = 0; i < 3; i++){
    b1[i] = a[i] - b[i];
    b2[i] = b[i] - c[i];
    b3[i] = c[i] - d[i];
  }

  /* add 20090819 */
  if(angle(b1, b2) * 180.0/PI > 175.0) return 999.0;
  if(angle(b2, b3) * 180.0/PI > 175.0) return 999.0;
  /* add end */

  cross(b1, b2, c1);
  cross(b2, b3, c2);
  cross(c1, c2, c3);

  torsion = angle(c1, c2);
/* 本当は回転の向きがある
  if (innerProduct(b2, c3) > 0.0) torsion = -torsion;
*/

  return torsion;
}

void cross(double *v1, double *v2, double *a){
  a[0] =   v1[1] * v2[2] - v1[2] * v2[1];
  a[1] = - v1[0] * v2[2] + v1[2] * v2[0];
  a[2] =   v1[0] * v2[1] - v1[1] * v2[0];
  return;
}

/* 角度：ラジアンで返す */
double angle(double *v1, double *v2){ 
  double ct;
  double v1len, v2len;

  v1len = sqrt(v1[0] * v1[0] + v1[1] * v1[1] + v1[2] * v1[2]);
  v2len = sqrt(v2[0] * v2[0] + v2[1] * v2[1] + v2[2] * v2[2]);
  if (v1len < NEARLY_ZERO || v2len < NEARLY_ZERO) return 999.0;

  ct = innerProduct(v1, v2) / (v1len * v2len);
  return acos(ct);
}

/* 内積 */
double innerProduct(double *v1, double *v2){
  return v1[0] * v2[0] + v1[1] * v2[1] + v1[2] * v2[2];
}

/* 結合の不一致情報を原子に移行(出力情報のため) */
void transBtoA(MCS *mcs, SIMCOMP *simcomp){
  int i;

  for (i = 0; i < simcomp->cpdmol_1->bondnum; i++){
    if (mcs->bondmap[i] < 0){
      if (mcs->atommap[simcomp->cpdmol_1->bond[i].a1 - 1] > 0)
        mcs->atommap[simcomp->cpdmol_1->bond[i].a1 - 1] *= -1;
      if (mcs->atommap[simcomp->cpdmol_1->bond[i].a2 - 1] > 0)
        mcs->atommap[simcomp->cpdmol_1->bond[i].a2 - 1] *= -1;
    }
  }
  return;
}

/*
  最大スコアの MCS を設定する
  スコアが同じときは、sccs のより大きな成分を含むものを勝ちとする
*/
void set_maxMCS(SIMCOMP *simcomp){
  int i, j;
  MCS_BLOCK *p;
  MCS *mcs;
  int lastnum;

  simcomp->max_mcs = NULL;

  if (simcomp->mcslist.num == 0){
    simcomp->status = 1;
    return;
  }

  p = simcomp->mcslist.pTop;

  while(p != NULL) {
    if (p->pNext != NULL){
      lastnum = 100;
    } else {
      lastnum = ((simcomp->mcslist.num - 1) % 100) + 1;
    }

    for (i = 0; i < lastnum; i++){
      mcs = &(p->mcs[i]);
      if (mcs->vnum == 0) continue;

      if (simcomp->max_mcs == NULL){  /* 未設定なら */
        simcomp->max_mcs = mcs;
      } else {
        if (simcomp->max_mcs->score < mcs->score){
          simcomp->max_mcs = mcs;
        } else
        if (simcomp->max_mcs->score == mcs->score){ /* sccsを比較 */
          for (j = 0; j < simcomp->max_mcs->sccsnum && j < mcs->sccsnum; j++){
            if (simcomp->max_mcs->sccs[j].vnum > mcs->sccs[j].vnum){
              break;
            } else
            if (simcomp->max_mcs->sccs[j].vnum < mcs->sccs[j].vnum){
              simcomp->max_mcs = mcs;
              break;
            }
          }
        }
      }
    }
    p = p->pNext;
  }

  /* trickによりMAX_MCSが消滅することがある 2008.08.18 */
  if (simcomp->max_mcs == NULL){
    simcomp->status = 1;
    return;
  }

  simcomp->mcs_max = simcomp->max_mcs->vnum;
#ifdef DEBUG
if (simcomp->debug_mode % 2 == 1){
printf("set mcs_max = %d\n", simcomp->mcs_max);
printf("ATOM MAP\n");
for (j = 0; j < simcomp->cpdmol_1->atomnum; j++){
printf("%2d - %2d\n", j + 1, simcomp->max_mcs->atommap[j]);
}
printf("vlist");
for (j = 0; j < simcomp->columnnum; j++){
printf(" %08X", simcomp->max_mcs->vlist[j]);
}
printf(" : score = %5.2f: sccs", simcomp->max_mcs->score);
for (j = 0; j < simcomp->max_mcs->sccsnum; j++){
printf(" (%d, %5.2f)", simcomp->max_mcs->sccs[j].vnum, simcomp->max_mcs->sccs[j].score);
}
printf("\n");
}
#endif
  return;
}

/* max 以外の MCS を inactive(vnum = 0) にする */
void set_inactive(SIMCOMP *simcomp){
  int i;
  MCS_BLOCK *p;
  MCS *mcs;
  int lastnum;

  if (simcomp->max_mcs == NULL) return;
  if (simcomp->mcslist.num == 0) return;

  p = simcomp->mcslist.pTop;

  while(p != NULL) {
    if (p->pNext != NULL){
      lastnum = 100;
    } else {
      lastnum = simcomp->mcslist.num % 100;
    }

    for (i = 0; i < lastnum; i++){
      mcs = &(p->mcs[i]);
      if (mcs->vnum == 0) continue;
      if (simcomp->max_mcs == mcs) continue;

      mcs->vnum = 0;
    }
    p = p->pNext;
  }

  return;
}

MCS_BLOCK* create_MCS_BLOCK(){
  MCS_BLOCK* p;

  p = (MCS_BLOCK *)malloc(sizeof(MCS_BLOCK));

  if (p != NULL) memset(p, 0, sizeof(MCS_BLOCK));

  return p;
}

void freeMember_MCS(MCS *p){
  int i;

  if (p == NULL) return;

  if (p->vlist != NULL){
    free(p->vlist);
    p->vlist = NULL;
  }
  if (p->sccs != NULL){
    for (i = 0; i < p->sccsnum; i++){
      if (p->sccs[i].vlist != NULL){
        free(p->sccs[i].vlist);
      }
    }
    free(p->sccs);
    p->sccs = NULL;
  }
  if (p->atommap != NULL){
    free(p->atommap);
    p->atommap = NULL;
  }
  if (p->bondmap != NULL){
    free(p->bondmap);
    p->bondmap = NULL;
  }

  return;
}

void free_MCS_BLOCK(MCS_BLOCK *p){
  int i;

  if (p != NULL){
    for (i = 0; i < 100; i++){
       if (p->mcs[i].vlist != NULL) freeMember_MCS(&p->mcs[i]);
    }
    free(p);
  }

  return;
}

int setlist_MCS(MCS_CTL *pctl, int vnum, unsigned int *vlist, double score){
  MCS* mcs = NULL;

  if (pctl->num % 100 == 0) {  /* 新規追加 */
    MCS_BLOCK* p = create_MCS_BLOCK();

    if (p == NULL) return _NO_;
    p->pPrev = pctl->pBottom;
    p->pNext = NULL;

    if (pctl->pBottom != NULL) pctl->pBottom->pNext = p; /* 前の次は自分 */
    if (pctl->pTop == NULL) pctl->pTop = p;              /* 空だったら自分を先頭 */
    pctl->pBottom = p;                                   /* 自分を最後 */

    mcs = &(pctl->pBottom->mcs[0]);
  } else {
    mcs = &(pctl->pBottom->mcs[pctl->num % 100]);
  }

  mcs->vnum = vnum;
  mcs->vlist = vlist;
  mcs->score = score;
  pctl->num++;

  /* SCCSの領域はまだ確保しない */
  mcs->sccs = NULL;

  return _YES_;
}

void freelist_MCS(MCS_CTL *pctl){
  MCS_BLOCK* p = pctl->pTop;      /* 先頭より */

  if (p == NULL) return;

  while(p != NULL) {
    MCS_BLOCK* pNext = p->pNext;  /* 次のポインタを記憶 */
    free_MCS_BLOCK(p);            /* 開放 */
    p = pNext;                 /* 次 */
  }

  /* リストコントロール構造体を初期化する */
  pctl->pTop    = NULL;
  pctl->pBottom = NULL;
  pctl->num     = 0;

  return;
}

