/*
  simcompp.c
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "simcomp.h"


/* private function */
void print_help(int);
void print_help_simcomp();
void print_help_simcompp();
void print_help_compsearch();
void print_help_simcompa();
void print_help_simcomps();
void print_help_all();
void print_version(char*);
void opt_check(SIMCOMP*, int, char**);
int exec_simcomp(SIMCOMP*, char*);
int exec_simcompp(SIMCOMP*, char*);
int exec_compsearch(SIMCOMP*, char*);
int exec_simcompa(SIMCOMP*, char*);
int exec_1to1(SIMCOMP*, FILE*);
int exec_simcomps(SIMCOMP*, char*);
void clear_simcomp(SIMCOMP*);
void standby_simcomp(SIMCOMP*);
void set_EntryName(char**, char*);
/* add 20090818 */
void CountAtomPenalty(SIMCOMP *);
/* add end */

void print_help(int match_mode){
  switch(match_mode){
  case _MA_ENTRYtoENTRY_: /* simcomp */
    print_help_simcomp();
    break;
  case _MA_KCFtoKCF_:     /* simcompp */
    print_help_simcompp();
    break;
  case _MA_ENTRYinDB_:    /* compsearch */
    print_help_compsearch();
    break;
  case _MA_RANGEinDB_:    /* simcompa */
    print_help_simcompa();
    break;
  case _MA_ENTRIEStoENTRIES_:    /* simcomps */
    print_help_simcomps();
    break;
  default:
    print_help_all();
    break;
  }

  return;
}
void print_help_simcomp(){
  printf("Summary : compute the compound similarity score & coefficient\n");
  printf("Usage  : simcomp [option]\n");
  printf("Usage  : simcomp -m e_to_e [option]\n");
  printf("Options\n");
  printf("file specifier:\n");
  printf("\t-i/--input  <file>  input file (stdin)\n");
  printf("\t-o/--output <file>  output all computational result (stdout)\n");
  printf("\n");
  printf("entry specifier:\n");
  printf("\t-e/--entry  <entry> entry id 1 (null)\n");
  printf("\t-f/--entry2 <entry> entry id 2 (null)\n");
  printf("\n");
  printf("constraint specifier:\n");
  printf("\t-a/--atom <mode>    docking by atom strictly, set 0, 1 or 2(2)\n");
  printf("\t--trickatom <mode>  atom mode in extending, set 0, 1 or 2\n");
  printf("\t-d/--dnode <mode>   docking node, set 0, 1 or 2(0)\n");
  printf("\n");
  printf("output mode specifier:\n");
  printf("\t--trick <mode>      execute some tricks, set 0 or 2\n");
  printf("\t--kcfoutput         output in the KCF format\n");
  printf("\t--nochiral          skip chirality check\n");
  printf("\n");
  printf("specifier for experts:\n");
  printf("\t--repeat <integer>  repeat number for clifinder()\n");
  printf("\t-l/--limit <number> limit of docking node (none)\n");
  printf("\n");
  printf("\t--global            same as \'-d 2 -a 2 --trick 2 --trickatom 0\'\n");
  printf("\t--local             same as \'-d 2 -a 2 --trick 0 --trickatom 1\'\n");
  printf("\t--help              print help\n");
  printf("\t--version           print version\n");

  return;
}

void print_help_simcompp(){
  printf("Summary : compute the compound similarity score & coefficient\n");
  printf("Usage  : simcompp [option] <file1> <file2>\n");
  printf("Usage  : simcomp -m k_to_k [option] <file1> <file2>\n");
  printf("Options\n");
  printf("file specifier:\n");
  printf("\t-o/--output <file>  output all computational result (stdout)\n");
  printf("\n");
  printf("constraint specifier:\n");
  printf("\t-a/--atom <mode>    docking by atom strictly, set 0, 1 or 2(2)\n");
  printf("\t--trickatom <mode>  atom mode in extending, set 0, 1 or 2\n");
  printf("\t-d/--dnode <mode>   docking node, set 0, 1 or 2(0)\n");
  printf("\n");
  printf("output mode specifier: \n");
  printf("\t--trick <mode>      execute some tricks, set 0 or 2\n");
  printf("\t--kcfoutput         output in the KCF format\n");
  printf("\t--nochiral          skip chirality check\n");
  printf("\n");
  printf("specifier for experts: \n");
  printf("\t--repeat <integer>  repeat number for clifinder()\n");
  printf("\t-l/--limit <number> limit of docking node (none)\n");
  printf("\t--global            same as \'-d 2 -a 2 --trick 2 --trickatom 0\'\n");
  printf("\t--local             same as \'-d 2 -a 2 --trick 0 --trickatom 1\'\n");
  printf("\n");
  printf("\t--help              print help\n");
  printf("\t--version           print version\n");

  return;
}

void print_help_compsearch(){
  printf("Summary : search a compound database for similar compounds\n");
  printf("Usage  : compsearch [option] <database> <query>\n");
  printf("Usage  : simcomp -m 1_in_db [option] <database> <query>\n");
  printf("Options\n");
  printf("constraint specifier:\n");
  printf("\t-a/--atom <mode>    docking by atom strictly, set 0, 1 or 2(2)\n");
  printf("\t--trickatom <mode>  atom mode in extending, set 0, 1 or 2\n");
  printf("\t-d/--dnode <mode>   docking node, set 0, 1 or 2(0)\n");
  printf("\n");
  printf("output specifier:\n");
  printf("\t--trick <mode>      execute some tricks, set 0 or 2\n");
  printf("\t--kcfoutput         output in the KCF/RPAIR format\n");
  printf("\t--nochiral          skip chirality check\n");
  printf("\n");
  printf("specifier for experts:\n");
  printf("\t--repeat <integer>  repeat number for clique find\n");
  printf("\t-t <number>         threshold score (0.4)\n");
  printf("\t-l/--limit <number> limit of docking node (none)\n");
  printf("\t--global            same as \'-d 2 -a 2 --trick 2 --trickatom 0\'\n");
  printf("\t--local             same as \'-d 2 -a 2 --trick 0 --trickatom 1\'\n");
  printf("\n");
  printf("\t--help              print help\n");
  printf("\t--version           print version\n");

  return;
}

void print_help_simcompa(){
  printf("Summary : compute the compound similarity score & coefficient\n");
  printf("Usage  : simcompa [option]\n");
  printf("Usage  : simcomp -m in_db [option]\n");
  printf("Options\n");
  printf("file specifier:\n");
  printf("\t-i/--input  <file>  input file (stdin)\n");
  printf("\t-o/--output <file>  output all computational result (stdout)\n");
  printf("\n");
  printf("entry specifier:\n");
  printf("\t-e/--entry  <entry> entry id for the first entry\n");
  printf("\t-f/--entry2 <entry> entry id for the last entry\n");
  printf("\n");
  printf("constraint specifier:\n");
  printf("\t-a/--atom <mode>    docking by atom strictly, set 0, 1 or 2(2)\n");
  printf("\t--trickatom <mode>  atom mode in extending, set 0, 1 or 2\n");
  printf("\t-d/--dnode <mode>   docking node, set 0, 1 or 2(0)\n");
  printf("\n");
  printf("output mode specifier:\n");
  printf("\t--trick <mode>      execute some tricks, set 0 or 2\n");
  printf("\t--kcfoutput         output in the KCF format\n");
  printf("\t--nochiral          skip chirality check\n");
  printf("\n");
  printf("specifier for experts:\n");
  printf("\t--repeat <integer>  repeat number for clifinder()\n");
  printf("\t-l/--limit <number> limit of docking node (none)\n");
  printf("\t--global            same as \'-d 2 -a 2 --trick 2 --trickatom 0\'\n");
  printf("\t--local             same as \'-d 2 -a 2 --trick 0 --trickatom 1\'\n");
  printf("\n");
  printf("\t--help              print help\n");
  printf("\t--version           print version\n");

  return;
}

void print_help_simcomps(){
  printf("Summary : compute the compound similarity score & coefficient\n");
  printf("Usage  : simcomps [option] -E <file1> -F <file2> -i <DB file>\n");
  printf("Usage  : simcomp -m es_to_es [option]\n");
  printf("Options\n");
  printf("file specifier:\n");
  printf("\t-i/--input  <file>  DB file (stdin)\n");
  printf("\t-o/--output <file>  output all computational result (stdout)\n");
  printf("\n");
  printf("entry specifier:\n");
  printf("\t-E/--entry-list  <file> filename for the first entries\n");
  printf("\t-F/--entry-list2 <file> filename for the second entries\n");
  printf("\n");
  printf("constraint specifier:\n");
  printf("\t-a/--atom <mode>    docking by atom strictly, set 0, 1 or 2(2)\n");
  printf("\t--trickatom <mode>  atom mode in extending, set 0, 1 or 2\n");
  printf("\t-d/--dnode <mode>   docking node, set 0, 1 or 2(0)\n");
  printf("\n");
  printf("output mode specifier:\n");
  printf("\t--trick <mode>      execute some tricks, set 0 or 2\n");
  printf("\t--kcfoutput         output in the KCF format\n");
  printf("\t--nochiral          skip chirality check\n");
  printf("\n");
  printf("specifier for experts:\n");
  printf("\t--repeat <integer>  repeat number for clifinder()\n");
  printf("\t-l/--limit <number> limit of docking node (none)\n");
  printf("\t--global            same as \'-d 2 -a 2 --trick 2 --trickatom 0\'\n");
  printf("\t--local             same as \'-d 2 -a 2 --trick 0 --trickatom 1\'\n");
  printf("\n");
  printf("\t--help              print help\n");
  printf("\t--version           print version\n");

  return;
}

void print_help_all(){
  printf("Summary : compute the compound similarity score & coefficient\n");
  printf("Usage  : simcomp [option ...] [<file> etc..]\n");
  printf("Options\n");
  printf("matchingspecifier:\n");
  printf("\t-m/--match <mode>   matching mode\n");
  printf("\t                      e_to_e | entry_to_entry       (simcomp mode)\n");
  printf("\t                      k_to_k | kcf_to_kcf           (simcompp mode)\n");
  printf("\t                      1_in_db | trial               (compsearch mode)\n");
  printf("\t                      inner_db | round_robin        (simcompa mode)\n");
  printf("\t                      es_to_es | entries_to_entries (simcomps mode)\n");
  printf("\n");
  printf("file specifier:\n");
  printf("\t-i/--input  <file>  input file (stdin)\n");
  printf("\t                      (simcomp, simcompa, simcomps mode)\n");
  printf("\t-o/--output <file>  output all computational result (stdout)\n");
  printf("\t                      (simcomp, simcompp, simcompa, simcomps mode)\n");
  printf("\n");
  printf("entry specifier:\n");
  printf("\t-e/--entry  <entry> entry id 1 (null)\n");
  printf("\t                      (simcomp, simcompa mode)\n");
  printf("\t-f/--entry2 <entry> entry id 2 (null)\n");
  printf("\t                      (simcomp, simcompa mode)\n");
  printf("\t-E/--entry-list  <file> filename for the first entries\n");
  printf("\t                      (simcomps mode)\n");
  printf("\t-F/--entry-list2 <file> filename for the second entries\n");
  printf("\t                      (simcomps mode)\n");
  printf("\n");
  printf("constraint specifier:\n");
  printf("\t-a/--atom <mode>    docking by atom strictly, set 0, 1 or 2(2)\n");
  printf("\t--trickatom <mode>  atom mode in extending, set 0, 1 or 2\n");
  printf("\t-d/--dnode <mode>   docking node, set 0, 1 or 2(0)\n");
  printf("\n");
  printf("output mode specifier:\n");
  printf("\t--trick <mode>      execute some tricks, set 0 or 2\n");
  printf("\t--kcfoutput         output in the KCF format\n");
  printf("\t--nochiral          skip chirality check\n");
  printf("\n");
  printf("specifier for experts:\n");
  printf("\t--repeat <integer>  repeat number for clifinder()\n");
  printf("\t-t <number>         threshold score (0.4)\n");
  printf("\t                      (compsearch mode)\n");
  printf("\t-l/--limit <number> limit of docking node (none)\n");
  printf("\t--global            same as \'-d 2 -a 2 --trick 2 --trickatom 0\'\n");
  printf("\t--local             same as \'-d 2 -a 2 --trick 0 --trickatom 1\'\n");
  printf("\n");
  printf("\t--help              print help\n");
  printf("\t--version           print version\n");

  return;
}

void print_version(char *name){
  printf("%s: %s ver %s (build %s)\n", name, PACKAGE, VERSION, BUILD_ID);

  return;
}

int main(int argc, char **argv){
  int ret = 0;
  SIMCOMP simcomp;
  char *progname;
  int  namelen;

  /* 計算オプション初期値設定 */
  memset(&simcomp, 0, sizeof(SIMCOMP));
#ifdef DEBUG
  simcomp.debug_mode = 0;
#endif
  simcomp.help_mode = _HE_EXECUTE_;
  simcomp.atom_mode = _DC_NORMAL_;
/* modify 20090904
  simcomp.trick_atom = -1;
*/
  simcomp.trick_atom = _DC_ATOM_;
  simcomp.exec_mode = _EX_NORMAL_;
  simcomp.output_mode = _OS_NORMAL_;
  simcomp.chiral = _CC_BOTH_;
  simcomp.node_mode = _NT_BOND_;
  simcomp.maxrepeat = -1;
  simcomp.limit = 0;
  simcomp.threshold = -1.0;

  /* 引数の解釈　メモリ管理上全面的に書き直し*/
  /* getopt_long()を使用するかと考えたが、入力ファイル指定場所の自由度ため順次処理 */
  opt_check(&simcomp, argc, argv);

/* comment out 20090904
  if (simcomp.trick_atom == -1) simcomp.trick_atom = simcomp.atom_mode;
*/

  progname = argv[0];
  namelen = strlen(progname);
  if (namelen >= 8 && strcmp(&progname[namelen - 8], "simcompp") == 0){
    simcomp.match_mode = _MA_KCFtoKCF_;
  } else
  if (namelen >= 10 && strcmp(&progname[namelen - 10], "compsearch") == 0){
    simcomp.match_mode = _MA_ENTRYinDB_;
  } else
  if (namelen >= 8 && strcmp(&progname[namelen - 8], "simcompa") == 0){
    simcomp.match_mode = _MA_RANGEinDB_;
  } else
  if (namelen >= 8 && strcmp(&progname[namelen - 8], "simcomps") == 0){
    simcomp.match_mode = _MA_ENTRIEStoENTRIES_;
  }

  if (simcomp.help_mode == _HE_HELP_){
    print_help(simcomp.match_mode);
    clear_simcomp(&simcomp);
    exit(0);
  }
  if (simcomp.help_mode == _HE_VERSION_){
    print_version(argv[0]);
    clear_simcomp(&simcomp);
    exit(0);
  }

  switch(simcomp.match_mode){
  case _MA_ENTRYtoENTRY_: /* simcomp */
    ret = exec_simcomp(&simcomp, argv[0]);
    break;
  case _MA_KCFtoKCF_:     /* simcompp */
    ret = exec_simcompp(&simcomp, argv[0]);
    break;
  case _MA_ENTRYinDB_:    /* compsearch */
    ret = exec_compsearch(&simcomp, argv[0]);
    break;
  case _MA_RANGEinDB_:    /* simcompa */
    ret = exec_simcompa(&simcomp, argv[0]);
    break;
  case _MA_ENTRIEStoENTRIES_:    /* simcomps */
    ret = exec_simcomps(&simcomp, argv[0]);
    break;
  default:                /* simcomp(default) */
    ret = exec_simcomp(&simcomp, argv[0]);
    break;
  }

  /* クロージング */
  clear_simcomp(&simcomp);
  exit(ret);
}

int exec_simcomps(SIMCOMP *simcomp, char *name){
  int ret = 0;
  int i = 0;

/* DBファイル中の最大登録データ数 */
#define DBNUM 1024

  FILE *input_file = stdin;
  FILE *output_file = stdout;
  FILE *entrylist1_file = NULL;
  FILE *entrylist2_file = NULL;
//  CPDMOL *mol = NULL;
  char	entryname[2][DBNUM][32];
  CPDMOL *mollist[DBNUM];
  int	list1num,list2num,molnum;
  int   j,k;

  /* マッチングモードに伴う無効オプションのクリア */
  if (simcomp->threshold == 0.0){
    printf("ERROR : Invalid option \"-t %lf\"\n", simcomp->threshold);
    simcomp->threshold = 0.0;
    return 0;
  }
  for (i = 0; i < simcomp->input_fn_num; i++){
    printf("ERROR : Invalid option \"%s\"\n", simcomp->input_fn[i]);
    return 0;
  }

   /*ファイルチェック */

  /* entry list file open */
  list1num = list2num = 0;
  if ( simcomp->entrylist1 != NULL ){
    if ((entrylist1_file = fopen(simcomp->entrylist1, "r")) == NULL){
      perror(simcomp->entrylist1);
      return 9;
    }
  }else{
    printf("Specify -E entry list file name\n" );
    return 0;
  }

  while ( fscanf(entrylist1_file,"%s",entryname[0][list1num]) != EOF ){
    if ( strlen(entryname[0][list1num]) != 0){
      list1num++;
      if ( list1num >= DBNUM ){
        printf("ID over %d\n",list1num );
        fclose(entrylist1_file);
        return 9;
      }
    }
  }
  fclose(entrylist1_file);

  if ( simcomp->entrylist2 != NULL ){
    if ((entrylist2_file = fopen(simcomp->entrylist2, "r")) == NULL){
      perror(simcomp->entrylist2);
      return 9;
    }
  }else{
    printf("Specify -F entry list file name\n" );
    return 0;
  }

  while ( fscanf(entrylist2_file,"%s",entryname[1][list2num]) != EOF ){
    if ( strlen(entryname[1][list2num]) != 0){
      list2num++;
      if ( list2num >= DBNUM ){
        printf("ID over %d\n",list2num );
        fclose(entrylist2_file);
        return 9;
      }
    }
  }
  fclose(entrylist2_file);

  /*DBファイルのオープン */
  if (simcomp->db_file != NULL){
    if ((input_file = fopen(simcomp->db_file, "r")) == NULL){
      perror(simcomp->db_file);
      return 9;
    }
  }else{
    printf("Specify -i Database file name\n" );
    return 0;
  }

  molnum=0;
  while((mollist[molnum] = read_CPDMOL_KCF(input_file, simcomp->chiral)) != NULL){
#ifdef DEBUG
    	printf("name =%s\n",mollist[molnum]->name);
#endif
    molnum++;
    if ( molnum >= DBNUM ){
      printf("DB over %d\n",molnum );
      fclose(input_file);
      return 9;
    }
  }
  fclose(input_file);

  /* 出力先のオープン */
  if (simcomp->output_fn != NULL){
    if ((output_file = fopen(simcomp->output_fn, "w")) == NULL){
      return 9;
    }
  }

  /* ２つのエントリリストでループする */
  for (i=0;i < list1num; i++){
    set_EntryName(&simcomp->entry1,entryname[0][i]);

    for(j=0;j < list2num; j++){
      set_EntryName(&simcomp->entry2,entryname[1][j]);

#ifdef DEBUG
    	printf("ent1 =%s,ent2 =%s\n",simcomp->entry1,simcomp->entry2);
#endif
      /* entry名の必須引数をチェック */
      if (simcomp->entry1 == NULL || simcomp->entry2 == NULL){
        printf("%s: specify two entry name\n", name);
        return 0;
      }

      simcomp->cpdmol_1 = simcomp->cpdmol_2 = NULL;
      /*** main procedure start ***/
      for (k=0; k< molnum ;k++){
        if (simcomp->cpdmol_1 == NULL &&
            strcasecmp(simcomp->entry1, mollist[k]->name) == 0){
          simcomp->cpdmol_1 = mollist[k];
          if (simcomp->cpdmol_2 != NULL) break;
        }/* else */
        if (simcomp->cpdmol_2 == NULL &&
            strcasecmp(simcomp->entry2, mollist[k]->name) == 0){
          simcomp->cpdmol_2 = mollist[k];
          if (simcomp->cpdmol_1 != NULL) break;
        }
      }

      if (simcomp->cpdmol_1 == NULL){
        printf("Skip %s: no such entry1(%s)\n", name, simcomp->entry1);
        continue;
      }
      if (simcomp->cpdmol_2 == NULL){
        printf("Skip %s: no such entry2(%s)\n", name, simcomp->entry2);
        continue;
      }

      ret = exec_1to1(simcomp, output_file);

      /*fprintf(output_file,"----------------------------------------------------\n");*/

      /* 最終の場合は、全体の処理をclear_simcompで行う */
      if (i != list1num-1 || j != list2num-1 ){
        standby_simcomp(simcomp);

      }

    }
  }

  fclose(output_file);

  /* MOLのクリア */
  for (k=0; k< molnum ;k++){
      /* clear_simcompでfreeするもの以外をfreeする。 */
    if ( mollist[k] != simcomp->cpdmol_1 && mollist[k] != simcomp->cpdmol_2){
      clear_CPDMOL(mollist[k]);
      mollist[k] = NULL;
    }
  }

  return ret;
}

int exec_simcomp(SIMCOMP *simcomp, char *name){
  int ret = 0;
  int i = 0;

  FILE *input_file = stdin;
  FILE *output_file = stdout;
  CPDMOL *mol = NULL;

  /* マッチングモードに伴う無効オプションのクリア */
  if (simcomp->threshold == 0.0){
    printf("ERROR : Invalid option \"-t %lf\"\n", simcomp->threshold);
    simcomp->threshold = 0.0;
    return 0;
  }
  for (i = 0; i < simcomp->input_fn_num; i++){
    printf("ERROR : Invalid option \"%s\"\n", simcomp->input_fn[i]);
    return 0;
  }

  /* entry名の必須引数をチェック */
  if (simcomp->entry1 == NULL || simcomp->entry2 == NULL){
    printf("%s: specify two entry name\n", name);
    return 0;
  }

  if (simcomp->db_file != NULL){
    if ((input_file = fopen(simcomp->db_file, "r")) == NULL){
      perror(simcomp->db_file);
      return 9;
    }
  }else{
    printf("Specify -i Database file name\n" );
    return 0;
  }

  if (simcomp->entrylist1 != NULL){
    printf("ERROR : Invalid option \"-E %s\"\n", simcomp->entrylist1);
    return 0;
  }
  if (simcomp->entrylist2 != NULL){
    printf("ERROR : Invalid option \"-F %s\"\n", simcomp->entrylist2);
    return 0;
  }

  /*** main procedure start ***/
  while((mol = read_CPDMOL_KCF(input_file, simcomp->chiral)) != NULL){
    if (simcomp->cpdmol_1 == NULL &&
        strcasecmp(simcomp->entry1, mol->name) == 0){
      simcomp->cpdmol_1 = mol;
      if (simcomp->cpdmol_2 != NULL) break;
    } else
    if (simcomp->cpdmol_2 == NULL &&
        strcasecmp(simcomp->entry2, mol->name) == 0){
      simcomp->cpdmol_2 = mol;
      if (simcomp->cpdmol_1 != NULL) break;
    } else {
      clear_CPDMOL(mol);
    }
    mol = NULL;
  }

  fclose(input_file);

  if (simcomp->cpdmol_1 == NULL){
    printf("%s: no such entry(%s)\n", name, simcomp->entry1);
    return 9;
  }
  if (simcomp->cpdmol_2 == NULL){
    printf("%s: no such entry(%s)\n", name, simcomp->entry2);
    return 9;
  }

  /* 出力先のオープン */
  if (simcomp->output_fn != NULL){
    if ((output_file = fopen(simcomp->output_fn, "w")) == NULL){
      return 9;
    }
  }

  ret = exec_1to1(simcomp, output_file);

  fclose(output_file);

  return ret;
}

int exec_simcompp(SIMCOMP *simcomp, char *name){
  int ret = 0;
  int i = 0;

  FILE *input_file = NULL;
  FILE *output_file = stdout;
#ifdef DEBUG
if (simcomp->debug_mode % 2 == 1){
printf("start exec_simcompp\n");
}
#endif

  /* KCFファイル名の必須引数をチェック */
  if (simcomp->input_fn_num < 2){
    printf("%s: specify only two input files\n", name);
    return 0;
  }
  for (i = 2; i < simcomp->input_fn_num; i++){
    printf("ERROR : Invalid option \"%s\"\n", simcomp->input_fn[i]);
    return 0;
  }

  /* マッチングモードに伴う無効オプションのクリア */
  if (simcomp->db_file != NULL){
    printf("ERROR : Invalid option \"-i %s\"\n", simcomp->db_file);
    return 0;
  }
  if (simcomp->entry1 != NULL){
    printf("ERROR : Invalid option \"-e %s\"\n", simcomp->entry1);
    return 0;
  }
  if (simcomp->entry2 != NULL){
    printf("ERROR : Invalid option \"-f %s\"\n", simcomp->entry2);
    return 0;
  }
  if (simcomp->threshold == 0.0){
    printf("ERROR : Invalid option \"-t %lf\"\n", simcomp->threshold);
    simcomp->threshold = 0.0;
    return 0;
  }
  if (simcomp->entrylist1 != NULL){
    printf("ERROR : Invalid option \"-E %s\"\n", simcomp->entrylist1);
    return 0;
  }
  if (simcomp->entrylist2 != NULL){
    printf("ERROR : Invalid option \"-F %s\"\n", simcomp->entrylist2);
    return 0;
  }

  /* ファイルの読み込み */
  if (simcomp->input_fn[0] != NULL){
    if((input_file = fopen(simcomp->input_fn[0], "r")) == NULL){
      perror(simcomp->input_fn[0]);
      return 9;
    }
  }else{
    printf("Specify input file name\n" );
    return 0;
  }

#ifdef DEBUG
if (simcomp->debug_mode % 2 == 1){
printf("call  read_CPDMOL_KCF1\n");
}
#endif
  simcomp->cpdmol_1 = read_CPDMOL_KCF(input_file, simcomp->chiral);
  fclose(input_file);

  /* ファイルの読み込み その２ */
  if (simcomp->input_fn[1] != NULL){
    if ((input_file = fopen(simcomp->input_fn[1], "r")) == NULL){
      return 9;
    }
  }else{
    printf("Specify input file name\n" );
    return 0;
  }

#ifdef DEBUG
if (simcomp->debug_mode % 2 == 1){
printf("call  read_CPDMOL_KCF2\n");
}
#endif
  simcomp->cpdmol_2 = read_CPDMOL_KCF(input_file, simcomp->chiral);
  fclose(input_file);

  /* 出力先のオープン */
  if (simcomp->output_fn != NULL){
    if ((output_file = fopen(simcomp->output_fn, "w")) == NULL){
      return 9;
    }
  }

  /*** main procedure start ***/
  ret = exec_1to1(simcomp, output_file);

  fclose(output_file);

  return ret;
}

int exec_compsearch(SIMCOMP *simcomp, char *name){
  int ret = 0;
  int i = 0;

  FILE *input_file = stdin;
  FILE *output_file = stdout;
  CPDMOL *mol = NULL;

  /* KCFファイル名の必須引数をチェック */
  if (simcomp->input_fn_num < 2){
    printf("%s: specify only two input files\n", name);
    return 0;
  }
  for (i = 2; i < simcomp->input_fn_num; i++){
    printf("ERROR : Invalid option \"%s\"\n", simcomp->input_fn[i]);
    return 0;
  }

  /* マッチングモードに伴う無効オプションのクリア */
  if (simcomp->db_file != NULL){
    printf("ERROR : Invalid option \"-i %s\"\n", simcomp->db_file);
    return 0;
  }
  if (simcomp->output_fn != NULL){
    printf("ERROR : Invalid option \"-o %s\"\n", simcomp->output_fn);
    return 0;
  }
  if (simcomp->entry1 != NULL){
    printf("ERROR : Invalid option \"-e %s\"\n", simcomp->entry1);
    return 0;
  }
  if (simcomp->entry2 != NULL){
    printf("ERROR : Invalid option \"-f %s\"\n", simcomp->entry2);
    return 0;
  }
  if (simcomp->entrylist1 != NULL){
    printf("ERROR : Invalid option \"-E %s\"\n", simcomp->entrylist1);
    return 0;
  }
  if (simcomp->entrylist2 != NULL){
    printf("ERROR : Invalid option \"-F %s\"\n", simcomp->entrylist2);
    return 0;
  }


  /* atom modeに応じたデフォルト値を設定 */
	/* modify 20090727 */
	/*
  if (simcomp->threshold <= 0.0 || simcomp->threshold > 1.0)
	*/
  if (simcomp->threshold < 0.0 || simcomp->threshold > 1.0){
	/* modify end */
    switch(simcomp->atom_mode){
    case _DC_NORMAL_:
      simcomp->threshold = _THRESHOLD_NID_;
      break;
    case _DC_CLASS_:
      simcomp->threshold = _THRESHOLD_CID_;
      break;
    case _DC_ATOM_:
      simcomp->threshold = _THRESHOLD_AID_;
      break;
    }
  }

  /* ファイルの読み込み */
  if((input_file = fopen(simcomp->input_fn[1], "r")) == NULL){
    perror(simcomp->input_fn[1]);
    return 9;
  }

  simcomp->cpdmol_1 = read_CPDMOL_KCF(input_file, simcomp->chiral);
  fclose(input_file);

  /* 出力先のオープン */
  if (simcomp->output_fn != NULL){
    if ((output_file = fopen(simcomp->output_fn, "w")) == NULL){
      return 9;
    }
  }

  if ((input_file = fopen(simcomp->input_fn[0], "r")) == NULL){
    perror(simcomp->input_fn[0]);
    return 9;
  }

  /*** main procedure start ***/
  while((mol = read_CPDMOL_KCF(input_file, simcomp->chiral)) != NULL){
    simcomp->cpdmol_2 = mol;
    ret = exec_1to1(simcomp, output_file);
    standby_simcomp(simcomp);

    clear_CPDMOL(mol);
    mol = NULL;
    simcomp->cpdmol_2 = NULL;
  }

  fclose(input_file);
  fclose(output_file);

  return 0;
}

int exec_simcompa(SIMCOMP *simcomp, char *name){
  int ret = 0;
  int i = 0;
  int finallap = 0;

  FILE *input_file = stdin;
  FILE *output_file = stdout;
  CPDMOL *mol = NULL;
  CPDMOL *lastmol = NULL;
  CPDMOL *startmol = NULL;
  CPDMOL *endmol = NULL;

  /* マッチングモードに伴う無効オプションのクリア */
  if (simcomp->threshold == 0.0){
    printf("ERROR : Invalid option \"-t %lf\"\n", simcomp->threshold);
    simcomp->threshold = 0.0;
    return 0;
  }
  for (i = 0; i < simcomp->input_fn_num; i++){
    printf("ERROR : Invalid option \"%s\"\n", simcomp->input_fn[i]);
    return 0;
  }
  if (simcomp->entrylist1 != NULL){
    printf("ERROR : Invalid option \"-E %s\"\n", simcomp->entrylist1);
    return 0;
  }
  if (simcomp->entrylist2 != NULL){
    printf("ERROR : Invalid option \"-F %s\"\n", simcomp->entrylist2);
    return 0;
  }

  /* 出力先のオープン */
  if (simcomp->output_fn != NULL){
    if ((output_file = fopen(simcomp->output_fn, "w")) == NULL){
      return 9;
    }
  }

  if (simcomp->db_file != NULL){
    if ((input_file = fopen(simcomp->db_file, "r")) == NULL){
      perror(simcomp->db_file);
      return 0;
    }
  }else{
    printf("Specify input file name\n" );
    return 0;
  }

  while((mol = read_CPDMOL_KCF(input_file, simcomp->chiral)) != NULL){
    if (lastmol == NULL){
      simcomp->mollist = mol;
      lastmol = mol;
    } else {
      lastmol->pNext = mol;
      mol->pPrev = lastmol;
      lastmol = mol;
    }

    if (startmol == NULL && simcomp->entry1 != NULL &&
        strcasecmp(simcomp->entry1, mol->name) == 0){
      if (endmol != NULL){  /* 選択順序が逆になっている */
        startmol = endmol;
        endmol = mol;
      } else {
        startmol = mol;
      }
    }
    if (endmol == NULL && simcomp->entry2 != NULL &&
        strcasecmp(simcomp->entry2, mol->name) == 0){
      endmol = mol;
    }
  }

  if (startmol == NULL) startmol = simcomp->mollist;
  if (endmol == NULL) endmol = lastmol;
  simcomp->cpdmol_1 = startmol;

  /*** main procedure start ***/
  while(simcomp->cpdmol_1 != NULL) {
    if (simcomp->cpdmol_1 == endmol) finallap = 1;
    simcomp->cpdmol_2 = simcomp->cpdmol_1->pNext;
    while(simcomp->cpdmol_2 != NULL){
      ret = exec_1to1(simcomp, output_file);
      standby_simcomp(simcomp);

      simcomp->cpdmol_2 = simcomp->cpdmol_2->pNext;
    }
    simcomp->cpdmol_1 = simcomp->cpdmol_1->pNext;
    if (finallap == 1) break;
  }

  fclose(input_file);
  fclose(output_file);

  return 0;
}

/* 1対1での計算 */
int exec_1to1(SIMCOMP *simcomp, FILE *output_file){
  double score;
#ifdef DEBUG
if (simcomp->debug_mode % 2 == 1){
printf("call check_nudonum\n");
}
#endif
  /* 分子のノード数をチェック */
  check_nudonum(simcomp);

  /* ドッキングノードを作成 */
  if (simcomp->status == 0){
    if (simcomp->exec_mode == _EX_NORMAL_ || simcomp->trick_atom == simcomp->atom_mode){
      alloc_DockingNode(simcomp);

      simcomp->nodenum = make_DockingNode(simcomp, simcomp->atom_mode, _YES_);
    } else {
      /* trick時のノードを先に生成　原子/結合の使用フラグを汚すため */
      alloc_DockingNode(simcomp);

      simcomp->nodenum = make_DockingNode(simcomp, simcomp->atom_mode, _YES_);
      simcomp->nodetrick_num = make_DockingNode(simcomp, simcomp->trick_atom, _NO_);
    }
  }
#ifdef DEBUG
if (simcomp->debug_mode % 2 == 1){
int i;
printf("==docking node==\n");
for (i = 0; i < simcomp->nodenum + simcomp->nodetrick_num; i++){
printf("DNODE(%d) Q(%d) - T(%d)\n", i + 1, simcomp->dnode[i].e1 + 1, simcomp->dnode[i].e2 + 1);
}
}
#endif

  /* 有効ノード数をチェック */
  if (simcomp->status == 0) check_effectiveNode(simcomp);

  /* compsearchの場合、thresholdでチェック */
  if (simcomp->status == 0){
    if (simcomp->match_mode == _MA_ENTRYinDB_){
      score = expect_similarity(simcomp);
      if (simcomp->threshold > score) return 0;
    }
  }

  /* ドッキンググラフを作成 */
  if (simcomp->status == 0) make_DockingGraph(simcomp);

  /* ビット処理用ワークを確保 */
  if (simcomp->status == 0){
    simcomp->bitset1 = (unsigned int*)malloc(sizeof(unsigned int) * simcomp->columnnum);
    simcomp->bitset2 = (unsigned int*)malloc(sizeof(unsigned int) * simcomp->columnnum);
    simcomp->bitset3 = (unsigned int*)malloc(sizeof(unsigned int) * simcomp->columnnum);
    simcomp->bitset4 = (unsigned int*)malloc(sizeof(unsigned int) * simcomp->columnnum);
    if (simcomp->bitset1 == NULL || simcomp->bitset2 == NULL ||
        simcomp->bitset3 == NULL || simcomp->bitset4 == NULL){
      return 9;
    }
    memset(simcomp->bitset1, 0, sizeof(unsigned int) * simcomp->columnnum);
    memset(simcomp->bitset2, 0, sizeof(unsigned int) * simcomp->columnnum);
    memset(simcomp->bitset3, 0, sizeof(unsigned int) * simcomp->columnnum);
    memset(simcomp->bitset4, 0, sizeof(unsigned int) * simcomp->columnnum);

    /* 原子とドッキングノードの対応関係 */
    /* 結合NODEで、KCF出力もしくはキラルチェックする時に使用 */
    if (simcomp->node_mode == _NT_BOND_ &&
        ((simcomp->output_mode & _OS_KCFFORMAT_) != 0 ||
          simcomp->chiral != _CC_NONE_)){
      simcomp->table1  = ctl_mallocArray_uint2(simcomp->cpdmol_1->atomnum, simcomp->columnnum);
      simcomp->table2  = ctl_mallocArray_uint2(simcomp->cpdmol_2->atomnum, simcomp->columnnum);
      if (simcomp->table1  == NULL || simcomp->table2  == NULL){
        return 9;
      }
    }
#ifdef DEBUG
if (simcomp->debug_mode % 2 == 1){
int i, j;
printf("==docking graph==\n");
for (i = 0; i < simcomp->nodenum + simcomp->nodetrick_num; i++){
printf("%d: ", i + 1);
for (j = 0; j < simcomp->columnnum; j++){
printf(" %08X", simcomp->matrix[i][j]);
}
printf("\n");
}
printf("==connection graph==\n");
for (i = 0; i < simcomp->nodenum + simcomp->nodetrick_num; i++){
printf("%d: ", i + 1);
for (j = 0; j < simcomp->columnnum; j++){
printf(" %08X", simcomp->connect[i][j]);
}
printf("\n");
}
}
#endif
  }

  /* 処理 */
  if (simcomp->status == 0){
    score = compute_similarity(simcomp);
  /* compsearchの場合、thresholdでチェック */
    if (simcomp->match_mode == _MA_ENTRYinDB_){
      if (simcomp->threshold > score) return 0;
    }

    /* add 20090819 */
    /* 原子のペナルティを設定する */
    CountAtomPenalty(simcomp);
    /* add end */
  }

#ifdef DEBUG
if (simcomp->debug_mode % 2 == 1){
printf("==output==\n");
}
#endif
  output_similarity(output_file, simcomp);

  fflush(output_file);

  return 0;
}

/*
　getopt_long()の使用を考えたが、
　必要なgetopt.hは多分あるだろうけど、
　入力ファイル指定場所の自由度がなくなるため、
　順次処理にて行う。
*/
void opt_check(SIMCOMP* simcomp, int argc, char** argv){
  int i;
  int global_local_flg = -1;

  simcomp->input_fn = (char **)malloc(sizeof(char *) * (argc - 1));
  if (simcomp->input_fn == NULL) return;
  memset(simcomp->input_fn, 0, sizeof(char *) * (argc - 1));

  for (i = 1; i < argc; i++) {
#ifdef DEBUG
    if (strcmp(argv[i], "--debug") == 0) {
      if (i + 1 < argc) {
        simcomp->debug_mode = atoi(argv[i + 1]);
        i++;
      }
    } else
#endif

    /* match specifier */
    if (strcmp(argv[i], "-m") == 0 ||
        strcmp(argv[i], "--match") == 0) {
      if (i + 1 < argc) {
        if (strcasecmp(argv[i + 1], "entry_to_entry") == 0 || strcasecmp(argv[i + 1], "e_to_e") == 0)
          simcomp->match_mode = _MA_ENTRYtoENTRY_;
        else
        if (strcasecmp(argv[i + 1], "kcf_to_kcf") == 0 || strcasecmp(argv[i + 1], "k_to_k") == 0)
          simcomp->match_mode = _MA_KCFtoKCF_;
        else
        if (strcasecmp(argv[i + 1], "1_in_db") == 0 || strcasecmp(argv[i + 1], "trial") == 0)
          simcomp->match_mode = _MA_ENTRYinDB_;
        else
        if (strcasecmp(argv[i + 1], "inner_db") == 0 || strcasecmp(argv[i + 1], "round_robin") == 0)
          simcomp->match_mode = _MA_RANGEinDB_;
        else
        if (strcasecmp(argv[i + 1], "entries_to_entries") == 0 || strcasecmp(argv[i + 1], "es_to_es") == 0)
          simcomp->match_mode = _MA_ENTRIEStoENTRIES_;
        i++;
      }
    } else
    /* file specifier */
    if (strcmp(argv[i], "-o") == 0 ||
        strcmp(argv[i], "--output") == 0) {
      if (i + 1 < argc) {
        StrAllocCopy(&simcomp->output_fn, argv[i + 1]);
        i++;
      }
    } else
    if (strcmp(argv[i], "-i") == 0 ||
        strcmp(argv[i], "--input") == 0) {
      if (i + 1 < argc) {
        StrAllocCopy(&simcomp->db_file, argv[i + 1]);
        i++;
      }
    } else
    /* entry specifier */
    if (strcmp(argv[i], "-e") == 0 ||
        strcmp(argv[i], "--entry") == 0) {
      if (i + 1 < argc) {
        set_EntryName(&simcomp->entry1, argv[i + 1]);
        i++;
      }
    } else
    if (strcmp(argv[i], "-f") == 0 ||
        strcmp(argv[i], "--entry2") == 0) {
      if (i + 1 < argc) {
        set_EntryName(&simcomp->entry2, argv[i + 1]);
        i++;
      }
    } else
    if (strcmp(argv[i], "-E") == 0 ||
        strcmp(argv[i], "--entry-list") == 0) {
      if (i + 1 < argc) {
        set_EntryName(&simcomp->entrylist1, argv[i + 1]);
        i++;
      }
    } else
    if (strcmp(argv[i], "-F") == 0 ||
        strcmp(argv[i], "--entry-list2") == 0) {
      if (i + 1 < argc) {
        set_EntryName(&simcomp->entrylist2, argv[i + 1]);
        i++;
      }
    } else
    /* constraint specifier */
    if (strcmp(argv[i], "-a") == 0 ||
        strcmp(argv[i], "--atom") == 0) {
      if(global_local_flg == -1){
        if (i + 1 < argc) {
          switch (atoi(argv[i + 1])) {
          case 0:
            simcomp->atom_mode = _DC_ATOM_;
            break;
          case 1:
            simcomp->atom_mode = _DC_CLASS_;
            break;
          case 2:
            simcomp->atom_mode = _DC_NORMAL_;
            break;
          default:
            simcomp->atom_mode = _DC_NORMAL_;
            break;
          }
          i++;
        }
      }
    } else
    if (strcmp(argv[i], "--trickatom") == 0) {
      if(global_local_flg == -1){
        if (i + 1 < argc) {
          switch (atoi(argv[i + 1])) {
          case 0:
            simcomp->trick_atom = _DC_ATOM_;
            break;
          case 1:
            simcomp->trick_atom = _DC_CLASS_;
            break;
          case 2:
            simcomp->trick_atom = _DC_NORMAL_;
            break;
          default:
            simcomp->trick_atom = -1;
            break;
          }
          i++;
        }
      }
    } else
    if (strcmp(argv[i], "-d") == 0 ||
        strcmp(argv[i], "--dnode") == 0) {
      if(global_local_flg == -1){
        if (i + 1 < argc) {
          switch (atoi(argv[i + 1])) {
          case 0:
            simcomp->node_mode = _NT_ATOM_;
            break;
          case 1:
            simcomp->node_mode = _NT_ATOM2_;
            break;
          case 2:
            simcomp->node_mode = _NT_BOND_;
            break;
          default:
            simcomp->node_mode = _NT_BOND_;
            break;
          }
          i++;
        }
      }
    } else
    /* output specifier */
    if (strcmp(argv[i], "--trick") == 0) {
      if(global_local_flg == -1){
        if (i + 1 < argc) {
          switch (atoi(argv[i + 1])) {
          case 0:
            simcomp->exec_mode = _EX_TRICK0_;
            break;
/*
          case 1:
            simcomp->exec_mode = _EX_TRICK1_;
            break;
*/
          case 2:
            simcomp->exec_mode = _EX_TRICK2_;
            break;
          default:
            simcomp->exec_mode = _EX_TRICK0_;
            break;
          }
          i++;
        }
      }
    } else
    if (strcmp(argv[i], "--global") == 0){
      /* -d 2 -a 2 --trick 2 --trickatom 0 */
      global_local_flg = 1;
      simcomp->node_mode = _NT_BOND_;
      simcomp->atom_mode = _DC_NORMAL_;
      simcomp->exec_mode = _EX_TRICK2_;
      simcomp->trick_atom = _DC_ATOM_;
   
    } else
    if (strcmp(argv[i], "--local") == 0){
      /* -d 2 -a 2 --trick 0 --trickatom 1 */
      global_local_flg = 2;
      simcomp->node_mode = _NT_BOND_;
      simcomp->atom_mode = _DC_NORMAL_;
      simcomp->exec_mode = _EX_TRICK0_;
      simcomp->trick_atom = _DC_CLASS_;
   
    } else
    if (strcmp(argv[i], "--kcfoutput") == 0) {
      simcomp->output_mode = _OS_KCFFORMAT_;
    /* add 20091021 start */
    } else
    if (strcmp(argv[i], "--kcfcolor2") == 0) {
      simcomp->output_mode = _OS_KCFFORMAT_COLOR2_;
    } else
    if (strcmp(argv[i], "--kcfcolor") == 0) {
      simcomp->output_mode = _OS_KCFFORMAT_COLOR_;
    /* add end */
    } else
    if (strcmp(argv[i], "--nochiral") == 0) {
      simcomp->chiral = _CC_NONE_;
    } else
    /* specifier for experts */
    if (strcmp(argv[i], "--repeat") == 0) {
      if (i + 1 < argc) {
        simcomp->maxrepeat = atoi(argv[i + 1]);
        i++;
      }
    } else
    if (strcmp(argv[i], "-t") == 0) {
      if (i + 1 < argc) {
        simcomp->threshold = strtod(argv[i + 1], NULL);
        i++;
      }
    } else
    if (strcmp(argv[i], "-l") == 0 ||
        strcmp(argv[i], "--limit") == 0) {
      if (i + 1 < argc) {
        simcomp->limit = atoi(argv[i + 1]);
        i++;
      }
    } else
    if (strcmp(argv[i], "--help") == 0) {
      simcomp->help_mode = _HE_HELP_;
    } else
    if (strcmp(argv[i], "--version") == 0) {
      simcomp->help_mode = _HE_VERSION_;
    } else
    if (strncmp(argv[i], "-", 1) == 0) {
      printf("ERROR : Illegal option \"%s\"\n", argv[i]);
      break;
    } else {
      /* DB, entry, file, etc.　マッチングモードに応じてあとで解釈する */
      StrAllocCopy(&simcomp->input_fn[simcomp->input_fn_num], argv[i]);
      simcomp->input_fn_num++;
    }
  }
}

/*
  エントリ名を補完する
-----単純なコピーに変更---------

  [Cc][0]nnn  ->  Cnnnnn
*/
void set_EntryName(char **dest, char *src){

    StrAllocCopy(dest, src);

/*-----単純なコピーに変更---------
  int i;

  if (strlen(src) > 6){
    StrAllocCopy(dest, src);
    return;
  }

  *dest = (char *)malloc(sizeof(char) * 7);

  strcpy(*dest, "C");
  if(*src == 'c' || *src == 'C') src++;
  if(strlen(src) < 5){
    for(i = 0; i < 5 - strlen(src); i++){
      strcat(*dest, "0");
    }
  }
  strcat(*dest, src);

-------------------------------*/

  return;
}

/* 終了後の後片付け */
void clear_simcomp(SIMCOMP *simcomp){
  int i;

  for (i = 0; i < simcomp->input_fn_num; i++){
    if (simcomp->input_fn[i] != NULL){
      free(simcomp->input_fn[i]);
    }
  }
  if (simcomp->input_fn != NULL){
    free(simcomp->input_fn);
  }
  if (simcomp->output_fn != NULL){
    free(simcomp->output_fn);
  }
  if (simcomp->db_file != NULL){
    free(simcomp->db_file);
  }
  if (simcomp->entry1 != NULL){
    free(simcomp->entry1);
  }
  if (simcomp->entry2 != NULL){
    free(simcomp->entry2);
  }


  if (simcomp->dnode != NULL){
    free(simcomp->dnode);
  }

  ctl_freeArray_uint2(&(simcomp->matrix), simcomp->nodenum + simcomp->nodetrick_num);
  ctl_freeArray_uint2(&(simcomp->connect), simcomp->nodenum + simcomp->nodetrick_num);
  freelist_MCS(&(simcomp->mcslist));
  if (simcomp->bitset1 != NULL){
    free(simcomp->bitset1);
  }
  if (simcomp->bitset2 != NULL){
    free(simcomp->bitset2);
  }
  if (simcomp->bitset3 != NULL){
    free(simcomp->bitset3);
  }
  if (simcomp->bitset4 != NULL){
    free(simcomp->bitset4);
  }
  if (simcomp->table1 != NULL){
    ctl_freeArray_uint2(&(simcomp->table1), simcomp->cpdmol_1->atomnum);
  }
  if (simcomp->table2 != NULL){
    ctl_freeArray_uint2(&(simcomp->table2), simcomp->cpdmol_2->atomnum);
  }

  if (simcomp->mollist != NULL){
    freelist_CPDMOL(&(simcomp->mollist));
  } else {
    if ( simcomp->cpdmol_1 != NULL ){
      /*　同じIDデータの場合の処理を追加 */
      if ( simcomp->cpdmol_1 != simcomp->cpdmol_2){
        clear_CPDMOL(simcomp->cpdmol_1);
      }
    }
    if ( simcomp->cpdmol_2 != NULL ){
      clear_CPDMOL(simcomp->cpdmol_2);
    }
  }
  return;
}

/* 次の実行のために一部クリア */
void standby_simcomp(SIMCOMP *simcomp){

  if (simcomp->dnode != NULL){
    free(simcomp->dnode);
    simcomp->dnode = NULL;
  }

  ctl_freeArray_uint2(&(simcomp->matrix), simcomp->nodenum + simcomp->nodetrick_num);
  ctl_freeArray_uint2(&(simcomp->connect), simcomp->nodenum + simcomp->nodetrick_num);

  simcomp->nodenum = 0;
  simcomp->columnnum = 0;
  simcomp->score = 0.0;
  simcomp->status = 0;
  simcomp->mcs_max = 0;
  simcomp->max_mcs = NULL;
  freelist_MCS(&(simcomp->mcslist));

  if (simcomp->bitset1 != NULL){
    free(simcomp->bitset1);
    simcomp->bitset1 = NULL;
  }
  if (simcomp->bitset2 != NULL){
    free(simcomp->bitset2);
    simcomp->bitset2 = NULL;
  }
  if (simcomp->bitset3 != NULL){
    free(simcomp->bitset3);
    simcomp->bitset3 = NULL;
  }
  if (simcomp->bitset4 != NULL){
    free(simcomp->bitset4);
    simcomp->bitset4 = NULL;
  }
  if (simcomp->table1 != NULL){
    ctl_freeArray_uint2(&(simcomp->table1), simcomp->cpdmol_1->atomnum);
  }
  if (simcomp->table2 != NULL){
    ctl_freeArray_uint2(&(simcomp->table2), simcomp->cpdmol_2->atomnum);
  }

  return;
}
