/*
　simcomp内の定義
*/


/*
　実行時オプション等　定数定義
*/
/* HElp mode */
#define _HE_EXECUTE_   (1u << 0) /* execute */
#define _HE_HELP_      (1u << 1) /* print help */
#define _HE_VERSION_   (1u << 2) /* print version */

/* MAtch mode */
#define _MA_ENTRYtoENTRY_   (1u << 0) /* simcomp    1_to_1(DB) */
#define _MA_KCFtoKCF_       (1u << 1) /* simcompp   1_to_1(KCF) */
#define _MA_ENTRYinDB_      (1u << 2) /* compsearch entry_in_DB */
#define _MA_RANGEinDB_      (1u << 3) /* simcompa   range_in_DB(round_robin) */
#define _MA_ENTRIEStoENTRIES_ (1u << 4) /* entriessearch m_to_n(DB) */

/* EXecution mode */
#define _EX_NORMAL_    (1u << 0) /* newstyle output mode */
#define _EX_TRICK0_    (1u << 2) /* post-extention and cutoff (mode 0) */
#define _EX_TRICK2_    (1u << 4) /* post-extention only       (mode 2) */

/* Output Style */
#define _OS_NORMAL_    (1u << 0) /* default output style */
#define _OS_KCFFORMAT_ (1u << 5) /* output in KCF */

/* add 20091021 start */
#define _OS_KCFFORMAT_COLOR_ (1u << 1) /* output in KCF */
#define _OS_KCFFORMAT_COLOR2_ (1u << 2) /* output in KCF */
/* add end */


/* Docking Constraint */
#define _DC_NORMAL_    (1u << 0) /* completely identical */
#define _DC_CLASS_     (1u << 1) /* class match */
#define _DC_ATOM_      (1u << 2) /* atom match */

/* Node Type */
#define _NT_ATOM_      (1u << 0) /* ノードを原子で作成 */
#define _NT_ATOM2_     (1u << 1) /* ノードを孤立してない原子で作成 */
#define _NT_BOND_      (1u << 2) /* ノードを結合で作成 */

/* Chiral Check */
#define _CC_NONE_      (1u << 0) /* キラルチェック：なし */
#define _CC_CHIRAL_    (1u << 1) /* キラルチェック：不斉 */
#define _CC_GEOMETRY_  (1u << 2) /* キラルチェック：trans/cis */
#define _CC_BOTH_      (_CC_CHIRAL_ | _CC_GEOMETRY_) /* キラルチェック：両方 */

#define _BIT_LENGTH_  32 /* unsigned intのビット数 */

#define _YES_          1
#define _NO_           0

#define _THRESHOLD_NID_  0.4  /* atom modeがアトムタイプの時の閾値(compsearch) */
#define _THRESHOLD_CID_  0.4  /* atom modeがクラスIDの時の閾値(compsearch) */
#define _THRESHOLD_AID_  0.4  /* atom modeが元素記号の時の閾値(compsearch) */

#define NEARLY_ZERO    0.0001  /* 絶対値がこれ以下だと0とみなす */
#define PI             3.14159265358979323846
#define PI_HALF        PI / 2.0

#define MAX_REPEAT 15000
#define BUILD_ID   "200910211130"

/*
　構造体
*/
/* 化合物ノード */
typedef struct _atom_ {
  char *nid;             /* nodetype NID */
  char *cid;             /* Class ID */
  char *aid;             /* atomの AID  */
  int HAflag;            /* H*かどうか？ 0:no 1:yes */
  int nH;                /* 水素との結合数 */
  int RAneighbor;        /* R*の近傍かどうか？ 0:no 1:yes */
  double x;              /* X座標 */
  double y;              /* Y座標 */
  char *prop1;           /* 原子のプロパティ1 */
  char *prop2;           /* 原子のプロパティ2 */
  int bondnum;           /* 結合数 */
  int *bond;             /* 結合の逆引き(0から数える) */
  int used;              /* グラフ作成時　：ドッキングノード生成に使用したか? 0:no 1:yes
                            クリーク探索時：extend_MCSの対象か? */
  int chiral;            /* キラルチェックの対象になるか? */
} ATOM;

typedef struct _bond_ {
  int a1;                /* atom1(1から数える) */
  int a2;                /* atom2(1から数える) */
  int order;             /* 結合次数 */
  int used;              /* グラフ作成時　：ドッキングノード生成に使用したか? 0:no 1:yes
                            クリーク探索時：extend_MCSの対象か? */
  int geometry;          /* trans/cisチェックの対象になるか? */
  char *prop;            /* 結合のプロパティ */
} BOND;

/* 化合物グラフ */
typedef struct _compound_mol_* LP_CPDMOL;
typedef struct _compound_mol_ {
  /* ファイルから読み込んだデータ */
  char *name;            /* entry 名 */
  char *definition;      /* definition 文字列 */
  int  atomnum;          /* 原子数 */
  int  atomnum_sub;      /* 原子数から RH* の原子を引いた数 */
  int  bondnum;          /* 結合数 */
  int  bondnum_sub;      /* 結合数から RH* の結合を引いた数 */
  ATOM *atom;            /* ATOM 情報 */
  BOND *bond;            /* BOND 情報 */
  int  nodenum;          /* score計算用(ドッキングノードの作成方法に依存しないように) */
  int  nodenum_sub;      /* score計算用(ドッキングノードの作成方法に依存しないように) */
  LP_CPDMOL pNext;       /* 次へのポインタ */
  LP_CPDMOL pPrev;       /* 前へのポインタ */
} CPDMOL;

/* ドッキングノード */
typedef struct _docking_node_ {
  int e1;                /* ノード番号１ */
  int e2;                /* ノード番号２ */
  double score;          /* ２ノード間のスコア */
  char ident;            /* ２ノードが完全一致かどうかのフラグ */
} DNODE;

/* SCCS */
typedef struct _graph_sccs_ {
  int vnum;              /* 頂点の数 */
  double score;          /* スコア */
  unsigned int *vlist;   /* 頂点のリスト(ビットで保持) */
} SCCS;

/* MCS */
typedef struct _graph_mcs_ {
  int vnum;              /* 頂点の数 */
  double score;          /* スコア */
  unsigned int *vlist;   /* 頂点のリスト(ビットで保持) */
  SCCS *sccs;
  double chiral_penalty; /* キラルチェックの結果によるスコアの減点分 */
  /* add 20090818 */
  double atom_penalty;   /* 原子タイプ、クラス、原子種毎によるスコアの減点分 */
  /* add end */
  int sccsnum;
  int *atommap;          /* 原子-原子マッピング(配列の添字は0から、値は1から数える) */
  int *bondmap;          /* 結合-結合マッピング(配列の添字は0から、値は1から数える) */
                         /* キラルチェックで不一致になると　マッピングを負の値にする */
} MCS;

/* 100個単位で管理する */
typedef struct _graph_mcsblock_* LP_MCS_BLOCK;
typedef struct _graph_mcsblock_ {
  MCS mcs[100];          /* MCSデータ */
  LP_MCS_BLOCK pNext;    /* 次へのポインタ */
  LP_MCS_BLOCK pPrev;    /* 前へのポインタ */
} MCS_BLOCK;

/* 出現数がわからないのでリストで管理 */
typedef struct _MCSL_CTL { /* リストコントロール構造体 */
  LP_MCS_BLOCK pTop;     /* 先頭へのポインタ */
  LP_MCS_BLOCK pBottom;  /* 末尾へのポインタ */
  int num;               /* データ数 */
} MCS_CTL;

/* SIMCOMP実行情報 */
typedef struct _simcomp_ {
  /* 実行時オプション */
#ifdef DEBUG
  int debug_mode;
#endif
  int help_mode;
  int match_mode;
  int atom_mode;
  int trick_atom;         /* trick時のatom_mode */
  int exec_mode;
  int node_mode;          /* ノードの作成方法 */
  int output_mode;
  int chiral;             /* キラリティチェック */
  int maxrepeat;
  int limit;
  double threshold;

  /* エントリー情報 */
  CPDMOL *cpdmol_1;       /* クエリー分子 */
  CPDMOL *cpdmol_2;       /* ターゲット分子 */
  CPDMOL *mollist;        /* simcompa用 */
  int input_fn_num;
  char **input_fn;        /* 入力ファイル名配列 */
  char *output_fn;        /* 出力ファイル名 */
  char *db_file;          /* DBファイル名 */
  char *entry1;           /* エントリーID1 */
  char *entry2;           /* エントリーID2 */
  char *entrylist1;       /* エントリーリスト1 */
  char *entrylist2;       /* エントリーリスト2 */

  /* 以下、実行するたびにクリアする */
  int nodenum;            /* ドッキングノード数 */
  int nodetrick_num;      /* ドッキングノード数(trick時に使用する分) */
  int columnnum;          /* ドッキングノードリストの数 */
  int alignnum;
  DNODE *dnode;           /* ドッキングノード情報(後半はtrick時に使用する分) */
  unsigned int **matrix;  /* 結合マトリックス */
  unsigned int **connect; /* SCCS作成用 行列 */
  double score;
  int status;             /* 計算結果のステータス */
  MCS_CTL mcslist;        /* MCSデータ */
  int mcs_max;
  MCS *max_mcs;           /* 最大スコアの MCS(実体の管理はmcslist) */
  unsigned int *bitset1;  /* ビット処理用ワーク1 */
  unsigned int *bitset2;  /* ビット処理用ワーク2 */
  unsigned int *bitset3;  /* ビット処理用ワーク3 */
  unsigned int *bitset4;  /* ビット処理用ワーク4 */
  unsigned int **table1;  /* 原子とドッキングノードの対応関係1 */
  unsigned int **table2;  /* 原子とドッキングノードの対応関係2 */
} SIMCOMP;

/* 計算結果のステータス */
/*
  0 ... 正常終了
  1 ... cpdmol のいずれかのノード数が 0 であった
  2 ... ドッキングノード数が 0 であった
  3 ... ドッキング時にノードオーバーフローを起こした　←変更に伴い発生しなくなった
  4 ... クリークオーバーフローを起こした(内部エラー)　←変更に伴い発生しなくなった
  5 ... 最大クリークのスコアが閾値に届かない
  6 ... メモリ確保に失敗
  9 ... cpdmol の何れかが NULL であった(入力エラー)
*/


/*
　関数プロトタイプ宣言
*/
/* simcompp.c */

/* cpdmol.c */
CPDMOL *read_CPDMOL_KCF(FILE*, int);
CPDMOL *cnode_CPDMOL(CPDMOL*);
void clear_CPDMOL(CPDMOL*);
void freelist_CPDMOL(CPDMOL**);

/* exe1.c */
double assume_similarity(SIMCOMP *);
double expect_similarity(SIMCOMP *);
double compute_similarity(SIMCOMP *);
void output_similarity(FILE *, SIMCOMP *);

unsigned int bit(int);
void setBit(unsigned int *, int);
void resetBit(unsigned int *, int);
void setBitInMatrix(unsigned int **, int, int);
unsigned int getBit(unsigned int *, int);
int bitCount(unsigned int);
unsigned int **ctl_mallocArray_uint2(int, int);
void ctl_freeArray_uint2(unsigned int ***, int);
char* StrAllocCopy(char**, char*);

/* ドッキンググラフ関係 */
/* cpddock.c */
int check_nudonum(SIMCOMP *);
void alloc_DockingNode(SIMCOMP *);
int make_DockingNode(SIMCOMP *, int, int);
int check_effectiveNode(SIMCOMP *);
void make_DockingGraph(SIMCOMP *);

int isAtomConnect(CPDMOL *, int, int);
int commonAtom(BOND *, BOND *);
int hasAtom(BOND *, int atom);

/* クリークファインド */
/* cpdcf2.c */
void clifinder2(SIMCOMP *);

/* MCS と SCCS 関係 */
/* mcs.c */
void score_MCS_list(SIMCOMP *);
void score_maxMCS(SIMCOMP *);
void make_MAP(MCS *, SIMCOMP *);
void set_maxMCS(SIMCOMP *);
void set_inactive(SIMCOMP *);

int setlist_MCS(MCS_CTL *, int, unsigned int *, double);
void freelist_MCS(MCS_CTL *);

